from epstein_files.documents.documents.categories import Neutral
from epstein_files.documents.config.config_builder import press_release
from epstein_files.documents.config.doc_cfg import DocCfg
from epstein_files.people.names import THE_REAL_DEAL
from epstein_files.util.constant.strings import (PALM_BEACH, PALM_BEACH_CODE_ENFORCEMENT, PALM_BEACH_TSV,
     PALM_BEACH_PROPERTY_INFO, PALM_BEACH_WATER_COMMITTEE, VIRGIN_ISLANDS)

REAL_DEAL_ARTICLE = 'article by Keith Larsen'


PROPERTY_CFGS = [
    DocCfg(id='016602', author=PALM_BEACH_CODE_ENFORCEMENT, note='board minutes', date='2008-04-17'),
    DocCfg(id='016554', author=PALM_BEACH_CODE_ENFORCEMENT, note='board minutes', date='2008-07-17', duplicate_ids=['016616', '016574']),
    DocCfg(id='016636', author=PALM_BEACH_WATER_COMMITTEE, note=f"Meeting on January 29, 2009"),
    DocCfg(id='022417', author='Park Partners NYC', note=f"letter to partners in real estate project with architectural plans"),
    DocCfg(id='027068', author=THE_REAL_DEAL, note=f"{REAL_DEAL_ARTICLE} Palm House Hotel Bankruptcy and EB-5 Visa Fraud Allegations"),
    DocCfg(id='029520', author=THE_REAL_DEAL, note=f"{REAL_DEAL_ARTICLE} 'Lost Paradise at the Palm House'", date='2019-06-17'),
    DocCfg(id='016597', author='Trump Properties LLC', note=f'appeal of some decision about Mar-a-Lago by {PALM_BEACH} authorities'),
    DocCfg(id='018743', note=f"Las Vegas property listing"),
    DocCfg(id='016695', note=f"{PALM_BEACH_PROPERTY_INFO} (?)"),
    DocCfg(id='016697', note=f"{PALM_BEACH_PROPERTY_INFO} (?) that mentions Trump"),
    DocCfg(id='016599', note=f"{PALM_BEACH_TSV} consumption (water?)"),
    DocCfg(id='016600', note=f"{PALM_BEACH_TSV} consumption (water?)"),
    DocCfg(id='016601', note=f"{PALM_BEACH_TSV} consumption (water?)"),
    DocCfg(id='016694', note=f"{PALM_BEACH_TSV} consumption (water?)"),
    DocCfg(id='016552', note=f"{PALM_BEACH_TSV} info"),
    DocCfg(id='016698', note=f"{PALM_BEACH_TSV} info (broken?)"),
    DocCfg(id='016696', note=f"{PALM_BEACH_TSV} info (water quality?"),
    DocCfg(
        id='018727',
        note=f"{VIRGIN_ISLANDS} property deal pitch deck, building will be leased to the U.S. govt GSA",
        date='2014-06-01',
    ),
    press_release('026759', 'Great Bay Condominium Owners Association', '2017-09-13', 'Hurricane Irma damage', is_interesting=False),
]



from epstein_files.documents.documents.categories import Interesting, Neutral, Uninteresting
from epstein_files.documents.doj_files.full_text import EFTA00009622_TEXT
from epstein_files.documents.config.communication_cfg import CommunicationCfg, Platform, imessage_log, imessage_screenshot, skype_log, whatsapp_log
from epstein_files.documents.config.config_builder import letter
from epstein_files.documents.config.doc_cfg import NO_TRUNCATE, DocCfg
from epstein_files.documents.config.email_cfg import EmailCfg
from epstein_files.output.highlight_config import DITE_ANATA_JUILLIARD_URL
from epstein_files.people.names import *
from epstein_files.util.constant.strings import *
from epstein_files.util.constant.urls import SVETLANA_NEWSGROUND
from epstein_files.util.helpers.string_helper import join_truthy, quote
from epstein_files.util.logging import logger

PRUSAKOVA_BERKELY = 'Epstein paid for Prusakova to go to Berkeley'
WOMEN_EMPOWERMENT = f"Women Empowerment (WE) conference"


GIRLS_CFGS = [
    DocCfg(id='019300', author=SVETLANA_POZHIDAEVA, note=f'{WOMEN_EMPOWERMENT} f. {KATHRYN_RUEMMLER}', date='2019-04-05', show_full_panel=True),
    DocCfg(id='022267', author=SVETLANA_POZHIDAEVA, note=f'{WOMEN_EMPOWERMENT} founder essay about growing the seminar business'),
    DocCfg(id='022407', author=SVETLANA_POZHIDAEVA, note=f'{WOMEN_EMPOWERMENT} seminar pitch deck'),
    EmailCfg(id='025874', note=f"possibly about {VIRGINIA_GIUFFRE} {QUESTION_MARKS}"),
    EmailCfg(id='030609', duplicate_ids=['030495']),

    # DOJ Files
    DocCfg(
        id='EFTA01334058',
        author='New York Post',
        category=Uninteresting.ARTICLE,
        date='2005-01-14',
        note="press snippets about Epstein's attempt to buy IMG Models as well as his relationships with Brett Ratner and Alina Pascau",
        is_in_chrono=False,
        is_interesting=True,
        truncate_to=1_200,
    ),
    DocCfg(
        id='EFTA01297974',
        note=f"DOJ queries OffshoreLeaks database for {SVETLANA_POZHIDAEVA}'s mother",
        show_full_panel=True,
        show_with_name=SVETLANA_POZHIDAEVA,
        truncate_to=500,
    ),
    DocCfg(id='EFTA02701609', author=GIANNI_SERAZZI, note='2morrow business plan', date='2014-09-30', date_uncertain=True),
    DocCfg(id='EFTA01078143', author=GIANNI_SERAZZI, note='2morrow modeling pitch deck', date='2014-09-30', date_uncertain=True),
    DocCfg(id='EFTA00306033', author=SERGEY_BELYAKOV, note='Epstein Russian visa', show_full_panel=True),
    DocCfg(
        id='EFTA01366289',
        author='Butterfly Trust',
        date='2015-03-30',
        is_interesting=20,
        note=f"transfer of $237,270 to {SVETLANA_POZHIDAEVA}'s father Raiffeisenbank from account at Standard Chartered Bank",
        show_full_panel=True,
        show_with_name=SVETLANA_POZHIDAEVA,
    ),
    DocCfg(id='EFTA00014648', note='former employee "understood that all the girls were school girls"', show_full_panel=True),
    DocCfg(id='EFTA00022133', note='table showing a list of Epstein accusations collection from public press reporting'),
    DocCfg(
        id='EFTA00077200',
        author=FRENCH_MINISTRY_OF_JUSTICE,
        date='2020-07-08',
        note=f'allegations that {JEAN_LUC_BRUNEL} drugged a model and Epstein + Ghislaine assaulted a minor',
    ),
    DocCfg(
        id='EFTA00079597',
        author=FRENCH_MINISTRY_OF_JUSTICE,
        note=f"allegations of sexual assault by {JEAN_LUC_BRUNEL}",
        date='2020-07-08',
    ),

    # Communications
    CommunicationCfg(
        id='EFTA01612665',
        note='Epstein gives advice on how to recruit girls',
        highlight_quote="stressed about finding girls. It's hard",
    ),
    imessage_log(
        id='EFTA02731576',
        is_interesting=10,
        note=f"making contemporaneous accusations",
        recipients=[LEON_BLACK],
        show_full_panel=True,
    ),
    imessage_screenshot('EFTA02731525', author=LEON_BLACK, author_uncertain=True),
    imessage_screenshot('EFTA01618381', author=RENATA_BOLOTOVA, author_uncertain='sneaky', highlight_quote='how many new girls are there in the island W you now?'),
    imessage_screenshot('EFTA01622387', author=RENATA_BOLOTOVA, author_uncertain='sneaky'),
    imessage_screenshot('EFTA01618494', author=RENATA_BOLOTOVA, author_uncertain='sneaky'),
    imessage_screenshot('EFTA01618400', author=RENATA_BOLOTOVA, author_uncertain='sneaky'),
    letter('EFTA00078198', 'Marsh Law Firm', [USANYS], "allegations of sexual abuse at Epstein's house", '2020-11-10'),
    skype_log(
        id='EFTA01616933',
        date='2019-04-08',
        note='"no one can beat your pussy network"',
        highlight_quote='now im finding pussy for you',
    ),
    skype_log(
        id='EFTA01623342',
        author=ANNA_KASATKINA,
        author_uncertain='https://www.reddit.com/r/Epstein/comments/1qwbn5i/trafficker_julia_santos/',
        note='recruiting russian girls',
        url='https://www.reddit.com/r/Epstein/comments/1qwbn5i/trafficker_julia_santos/',
    ),
    whatsapp_log(
        id='EFTA01612733',
        author=MARIA_PRUSAKOVA,
        author_uncertain=True,
        highlight_quote='way better than supplying you ladies',
        note=f'saying being Epstein\'s lawyer would be preferable to recruiting, discusses Medici Bank',
    ),

    # Authors
    EmailCfg(
        id='EFTA02339926',
        author=JULIA_SANTOS,
        author_uncertain='https://www.reddit.com/r/Epstein/comments/1qwbn5i/trafficker_julia_santos/',
        highlight_quote='very sweet and might be naughty t=o',
    ),
    EmailCfg(
        id='EFTA01745650',
        author=KIRA_DIKHTYAR,
        author_uncertain='AOL signature',
        note=f'{SANITA} is big mad at Epstein and "makes lists" which is freaking everyone out',
        show_with_name=SANITA,
        truncate_to=NO_TRUNCATE,
    ),
    EmailCfg(id='EFTA01781620', author=KIRA_DIKHTYAR, author_uncertain='"Sent from AOL Mobile Mail" in chain'),
    EmailCfg(id='EFTA01805304', author=KIRA_DIKHTYAR, author_reason='Jmail', truncate_to=NO_TRUNCATE),
    EmailCfg(id='EFTA01766199', author=KIRA_DIKHTYAR, author_reason='reply', truncate_to=NO_TRUNCATE),
    EmailCfg(
        id='033178',
        author=MARIA_PRUSAKOVA,
        author_reason=PRUSAKOVA_BERKELY,
        is_interesting=10,
        note=f'Masha Prusso asks about {ZUBAIR_KHAN}, discusses recruiting girls for Epstein',
    ),
    EmailCfg(
        id='EFTA00669538',
        author=MARIA_PRUSAKOVA,
        author_uncertain='discussion of "experiments"',
        note='blowjob instructions, sexual experiments',
        truncate_to=NO_TRUNCATE,
    ),
    EmailCfg(
        id='EFTA02010589',
        author=MARIA_PRUSAKOVA,
        date='2012-01-31 9:16 PM',
        note=f'raunchy email mentioning "Sasha Grey guy", note {BORIS_NIKOLIC} and bgC3 in SMTP info at end',
        is_interesting=10,
        recipients=[JEFFREY_EPSTEIN],
    ),
    EmailCfg(
        id='EFTA00928879',
        author=MARIA_PRUSAKOVA,
        author_uncertain='sounds like her',
        note=f'raunchy email',
        highlight_quote='your job is to find interesting girls',
        truncate_to=3_300,
    ),
    EmailCfg(
        id='EFTA01841553',
        author=MARIA_PRUSAKOVA,
        author_uncertain=True,
        highlight_quote='being fucked in the ass hurts me a lot',
        truncate_to=900,
    ),
    EmailCfg(
        id='032374',
        author=MARIA_PRUSAKOVA,
        author_reason=PRUSAKOVA_BERKELY,
        note='boyfriend named Christian mentioned',
        is_interesting=True,
    ),
    EmailCfg(id='032375', author=MARIA_PRUSAKOVA, author_reason=PRUSAKOVA_BERKELY, truncate_to=295),
    EmailCfg(id='033246', author=MARIA_PRUSAKOVA, author_reason='Kind regards/Cordialement in signature'),
    EmailCfg(id='EFTA01990879', author=MARIA_PRUSAKOVA, author_uncertain='Clifford Chance, could be Miranda'),
    EmailCfg(id='EFTA01990168', author=MARIA_PRUSAKOVA, author_uncertain='"experiments"', is_interesting=True),
    EmailCfg(id='EFTA00671662', author='Miranda', author_reason='quoted signature', note="yet another girl finder"),
    EmailCfg(id='EFTA01047249', author=NADIA_MARCINKO, author_reason='Miro, https://archive.ph/Qa6vU#selection-1621.160-1621.184'),
    EmailCfg(id='EFTA02441035', author=STEVEN_VICTOR_MD, note="complaints about free medical treatment for Epstein's girls"),
    EmailCfg(
        id='EFTA00659941',
        author=SVETLANA_POZHIDAEVA,
        author_reason='https://tommycarstensen.com/epstein/findings.html#victim-diary-names-multiple-men',
        is_interesting=10,
        note=f'{SVETLANA_POZHIDAEVA} forwarding her intimate conversations with {JOSHUA_FINK} to Epstein (to what end?)',
    ),
    EmailCfg(
        id='EFTA01844498',
        author=SVETLANA_POZHIDAEVA,
        author_reason='Josh Fink, BlackBerry signature',
        is_interesting=10,
        note=f"jealous boyfriend {JOSHUA_FINK} accuses Svetlana had an affair with Navalny {QUESTION_MARKS}",
    ),
    EmailCfg(
        id='EFTA01870453',
        author=SVETLANA_POZHIDAEVA,
        author_uncertain=f"{JOSHUA_FINK} texts",
        highlight_quote="You have known you are preg for a week",
        is_interesting=10,
        note=f'{SVETLANA_POZHIDAEVA} fwds a discussion about an abortion to Epstein',
    ),
    EmailCfg(
        id='EFTA01894879',
        note='"most of all its cheap excort"',
        duplicate_ids=['EFTA00876711'],
        highlight_quote='400 girls for model agencies and wedding agencies',
        recipients=[SVETLANA_POZHIDAEVA],
        recipient_uncertain='"Svet"',
    ),
    EmailCfg(
        id='EFTA01991293',
        author=SVETLANA_POZHIDAEVA,
        author_reason=f"convo about {JOSHUA_FINK} breakup",
        is_interesting=10,
        note=f'did {ELON_MUSK} date {SVETLANA_POZHIDAEVA} after she blackmailed {JOSHUA_FINK}? ("he asked 6 times if Elon [gave] me anything")',
    ),
    EmailCfg(
        id='EFTA01998247',
        author=SVETLANA_POZHIDAEVA,
        author_uncertain='device signature',
        note=f'trying to recruit {ELON_MUSK}?',
        is_interesting=10,
        show_with_name=ELON_MUSK,
    ),
    EmailCfg(
        id='EFTA01894132',
        author=SVETLANA_POZHIDAEVA,
        author_uncertain='"Svet" in unredacted reply',
        truncate_to=(900, 1_600),
    ),
    EmailCfg(id='EFTA01024246', note=f"invoice from {SVETLANA_POZHIDAEVA}'s WE Talks", show_with_name=SVETLANA_POZHIDAEVA),
    EmailCfg(id='EFTA02634573', author=SVETLANA_POZHIDAEVA, note='updating Epstein on Women Empowerment'),
    EmailCfg(id='EFTA01772677', author=SVETLANA_POZHIDAEVA, author_uncertain='device signature'),
    EmailCfg(id='EFTA01877224', author=SVETLANA_POZHIDAEVA, author_uncertain=f"{JOSHUA_FINK} texts", is_interesting=10),
    EmailCfg(id='EFTA00937507', author=SVETLANA_POZHIDAEVA, author_uncertain=f"{JOSHUA_FINK} texts", is_interesting=10),
    EmailCfg(
        id='EFTA01744405',
        author=SVETLANA_POZHIDAEVA,
        is_interesting=10,
        note=f"{SVETLANA_POZHIDAEVA} discusses her father (an FSB officer) and 'Josh' (presumably {JOSHUA_FINK})",
        url=SVETLANA_NEWSGROUND,
    ),
    EmailCfg(
        id='EFTA02475964',
        note=f"Epstein thanks {SVETLANA_POZHIDAEVA}'s father, a former (?) FSB officer",
        is_interesting=20,
        recipients=[SVETLANA_POZHIDAEVA],
        url=SVETLANA_NEWSGROUND,
    ),
    EmailCfg(
        id='EFTA02480401',
        highlight_quote='i re=iewed payment of 237,270 for svet family',
        is_interesting=20,
        note=f'"svet family" is {SVETLANA_POZHIDAEVA}\'s family in Russia',
        show_with_name=SVETLANA_POZHIDAEVA,
    ),
    EmailCfg(id='EFTA02431535', author=UNKNOWN_GIRL),
    EmailCfg(id='EFTA01875607', author=UNKNOWN_GIRL),
    EmailCfg(id='EFTA00901581', author=UNKNOWN_GIRL),
    EmailCfg(id='EFTA02449477', author=UNKNOWN_GIRL, note='argument about finding girls for Epstein', is_interesting=10, truncate_to=NO_TRUNCATE),
    EmailCfg(id='EFTA02027844', author=UNKNOWN_GIRL, note='permission to "fuck someone with a condom"', truncate_to=NO_TRUNCATE),
    EmailCfg(id='EFTA00763537', author=UNKNOWN_GIRL, highlight_quote="19 but I look younger than her"),
    EmailCfg(id='EFTA01782788', author=UNKNOWN_GIRL, highlight_quote='I need 21-24, wiling to travel and work hard', truncate_to=NO_TRUNCATE),
    EmailCfg(
        id='EFTA01024118',
        author=YULIA_DOROKHINA,
        author_reason='juliador89@mail.ru visible in verticle text',
        highlight_quote='I sent you nude pictures, sexy videos even masturbation',
        truncate_to=1_900,
    ),
    EmailCfg(
        id='EFTA01035876',
        author=YULIA_DOROKHINA,
        author_uncertain='"Yulia" in reply',
        note=f'looking for a job under {JIDE_ZEITLIN}',
        show_with_name=JIDE_ZEITLIN,
    ),
    EmailCfg(
        id='EFTA01036212',
        author=YULIA_DOROKHINA,
        author_uncertain='"Yulia"',
        note=f"Epstein wants {YULIA_DOROKHINA} to travel on short notice",
        truncate_to=NO_TRUNCATE,
    ),
    EmailCfg(id='EFTA01025400', author=YULIA_DOROKHINA),
    EmailCfg(id='EFTA01025734', author=YULIA_DOROKHINA, truncate_to=1_300),
    EmailCfg(id='EFTA01024680', author=YULIA_DOROKHINA, author_uncertain='thread'),
    EmailCfg(id='EFTA01024729', author=YULIA_DOROKHINA, author_uncertain='thread'),
    EmailCfg(id='EFTA02634523', author=YULIA_DOROKHINA, author_uncertain='"Yulia"'),
    EmailCfg(id='EFTA01028241', author=YULIA_DOROKHINA, author_uncertain='"Yulia" in other emails in thread', truncate_to=NO_TRUNCATE),
    EmailCfg(id='EFTA01026136', author=YULIA_DOROKHINA, highlight_quote='send the masturbation video', truncate_to=5000),

    # Recipients
    EmailCfg(
        id='EFTA02223525',
        recipients=[ANNA_KASATKINA, BELLA_KLEIN, JANUSZ_BANASIAK],
        recipient_uncertain='https://www.reddit.com/r/Epstein/comments/1qwbn5i/trafficker_julia_santos/',
        truncate_to=800
    ),
    EmailCfg(id='EFTA02470883', recipients=[YULIA_DOROKHINA]),
    EmailCfg(
        id='EFTA00997627',
        highlight_quote='you will be introduced to more gorgeous Alicas, i promise',
        recipients=[MARIA_PRUSAKOVA],
        recipient_uncertain='Kind regards/Cordialement in signature',
    ),
    EmailCfg(id='EFTA00719146', recipients=[MARIA_PRUSAKOVA], recipient_uncertain='Clifford Chance, could be Miranda'),
    EmailCfg(
        id='EFTA01857628',
        note='"igrushki" means "toys" in Russian',
        recipients=[RENATA_BOLOTOVA],
        recipient_uncertain=True,
        url='https://x.com/FlippersUpNow/status/2023457206109155345',
    ),
    EmailCfg(
        id='EFTA00857669',
        note=f"{SANITA} is maybe the only great girl with right thinking in the entire Epstein files",
        recipients=[SANITA],
        truncate_to=NO_TRUNCATE,
    ),
    EmailCfg(id='EFTA00738485', recipients=[STEVEN_VICTOR_MD], truncate_to=640),
    EmailCfg(
        id='EFTA02560884',
        is_interesting=10,
        note=f"is Epstein drafting {SVETLANA_POZHIDAEVA}'s romantic responses to {JOSHUA_FINK} {QUESTION_MARKS}",
        recipients=[SVETLANA_POZHIDAEVA],
        recipient_uncertain=f"{JOSHUA_FINK} texts",
    ),
    EmailCfg(
        id='EFTA02543909',
        author_reason='p.selena@yahoo.com in replies',
        is_interesting=10,
        note=f'{SVETLANA_POZHIDAEVA}\'s relationship with "J" ({JOSHUA_FINK}?) and "B" {QUESTION_MARKS}',
        recipients=[SVETLANA_POZHIDAEVA],
    ),
    EmailCfg(
        id='EFTA00937556',
        is_interesting=10,
        note=f"looks like Epstein coaching {SVETLANA_POZHIDAEVA} on how to respond to {JOSHUA_FINK}",
        recipients=[SVETLANA_POZHIDAEVA],
        recipient_uncertain=True,
    ),
    EmailCfg(
        id='EFTA02447019',
        url='https://www.wsj.com/us-news/epstein-victim-svetlana-pozhidaeva-22f2ebf6',
        recipients=[SVETLANA_POZHIDAEVA],
    ),
    EmailCfg(
        id='EFTA00888467',
        is_interesting=True,
        note=f'very strange conversation about "Peter" (maybe {PETER_MANDELSON}?) and "mj" (maybe {HASSAN_JAMEEL}?)',
        recipients=[UNKNOWN_GIRL],
    ),
    EmailCfg(id='EFTA00897668', recipients=[UNKNOWN_GIRL], entity_names=[JEAN_LUC_BRUNEL, JEFFREY_EPSTEIN, UNKNOWN_GIRL]),
    EmailCfg(id='EFTA00848644', recipients=[UNKNOWN_GIRL], highlight_quote='take a selfie of your pussy and send'),
    EmailCfg(
        id='EFTA01026268',
        author_reason='juliador89®mail.ru is unredacted',
        is_interesting=10,
        note=f"Epstein pressuring {YULIA_DOROKHINA} to find girls in Russia and send him nude pics and videos",
        recipients=[YULIA_DOROKHINA],
        truncate_to=3_100,
    ),
    EmailCfg(id='EFTA01024839', recipients=[YULIA_DOROKHINA], recipient_uncertain='part of thread'),
    EmailCfg(id='EFTA02276109', recipients=[YULIA_DOROKHINA], recipient_uncertain='"Yulia"'),

    # Descriptions
    EmailCfg(id='EFTA00078209', note='allegations about Russian organized crime, sex trafficking', truncate_to=8_500),
    EmailCfg(id='EFTA01648951', note=f'allegations against Trump, {HOWARD_LUTNICK}, Glen Dubin, {LEON_BLACK}, {JES_STALEY}'),
    EmailCfg(id='EFTA00950368', note='"any [girl] friends for me?"', truncate_to=NO_TRUNCATE),
    EmailCfg(id='EFTA01814983', note=f"Epstein and {NICOLE_JUNKERMANN} used to date", show_with_name=NICOLE_JUNKERMANN),
    EmailCfg(id='EFTA02049364', note='Epstein arranging for healthcare for five girls', truncate_to=NO_TRUNCATE),
    EmailCfg(id='EFTA02209622', note=f'flight to Riga for {LASMA_KUHTARSKA}', show_with_name=LASMA_KUHTARSKA, truncate_to=600),
    EmailCfg(id='EFTA00968086', note='request for help for a Juilliard student', url=DITE_ANATA_JUILLIARD_URL),
    EmailCfg(
        id='EFTA00883738',
        note=f'Epstein recruiter {DASHA_GRUPMAN} looking to spend $50,000 to rent a townhouse for "diplomats"',
        show_with_name=DASHA_GRUPMAN,
    ),
    DocCfg(
        id='EFTA00009622',
        date='2006-07-19',
        note='handwritten notes from a victim interview transcribed by Claude AI',
        is_interesting=True,
        display_text=EFTA00009622_TEXT,
    ),
    EmailCfg(
        id='EFTA00577409',
        date='2003-06-01',
        date_uncertain='just a guess, string is 4501 which is NULL in Microsoft Outlook',
        note='Epstein and Ghislaine discuss girl "from donad trumps party" who blackmailed [someone]',
    ),
    EmailCfg(
        id='EFTA02007083',
        note=f"{BORIS_NIKOLIC} encourages {KIMBAL_MUSK} to ditch his girlfriend for Epstein's girl {JENNIFER_KALIN} (and he says yes)",
        truncate_to=NO_TRUNCATE,
    ),
    EmailCfg(
        id='EFTA00412260',
        author=LESLEY_GROFF,
        comment='very broken file',
        date='2012-05-25 19:51:20',
        note=f'emphasis quotes by {LESLEY_GROFF}',
        highlight_quote='Jeffrey was asking if you and "your friend" could come by and meet Peter',
        recipients=[None],
        truncate_to=(6_700, 7_450),
    ),
    EmailCfg(
        id='EFTA00438431',
        note=f"visa invitation for {KARYNA_SHULIAK} to visit {QUESTION_MARKS}",
        recipients=[LESLEY_GROFF],
        show_with_name=KARYNA_SHULIAK,
    ),

    # Quotes
    EmailCfg(
        id='EFTA02504370',
        highlight_quote=f'lovely girl with financial troubles',
        recipients=[KIRA_DIKHTYAR],
        recipient_uncertain='"Sent from AOL Mobile Mail"',
        truncate_to=NO_TRUNCATE,
    ),
    EmailCfg(
        id='EFTA01022353',
        highlight_quote='she said she was 14-15 yo',
        note='(George Models is basically a porn site)',
        is_interesting=True,
    ),
    EmailCfg(
        id='EFTA01885139',
        highlight_quote='Kimbal really like Jen',
        is_interesting=True,
        note=f'({JENNIFER_KALIN}, Epstein\'s girl whom {KIMBAL_MUSK} soon dated)',
    ),
    EmailCfg(
        id='EFTA02415420',
        highlight_quote="I have a friend of Putin,s",
        is_interesting=10,
    ),
    EmailCfg(
        id='EFTA02555750',
        highlight_quote='suspend any billing from now on until further notification from you or the model',
        note=f"who's \"the model\" {JEAN_LUC_BRUNEL} is billing {JOSHUA_FINK} for?",
        is_interesting=10,
    ),
    EmailCfg(id='EFTA00971793', highlight_quote='Attached is resume of Lutnick nanny', note=f"why does Epstein need a nanny?"),
    EmailCfg(id='EFTA01965732', highlight_quote='facilictating his illicit trysts, with married women, to being asked to provide adderall fro bridge tournamnts'),
    EmailCfg(id='EFTA01768670', highlight_quote='find girls for the agency', note=f'but {BORIS_NIKOLIC} is a "biotech investor"...'),
    EmailCfg(id='EFTA01905320', highlight_quote='girls and i are going to see elon musk at space x tomorrow', is_interesting=True),
    EmailCfg(id='EFTA01803353', highlight_quote='having problems with regina', note='(Ramsey?)'),
    EmailCfg(id='EFTA01858707', highlight_quote='How is the marriage plan going'),
    EmailCfg(id='EFTA00950394', highlight_quote='how old will you go?'),
    EmailCfg(id='EFTA02499884', highlight_quote='I gave another girl to kimball and he is thrilled', truncate_to=NO_TRUNCATE),
    EmailCfg(id='EFTA02557757', highlight_quote="I have a girl Sana working half time/who by the way you'd like"),
    EmailCfg(id='EFTA00771856', highlight_quote="I know 23 is on the old side for you", is_interesting=True),
    EmailCfg(id='EFTA01974447', highlight_quote='I know you are going to meet putin on the 20th', is_interesting=True),
    EmailCfg(id='EFTA01767237', highlight_quote='is 24 too old for you?', truncate_to=NO_TRUNCATE),
    EmailCfg(id='EFTA00854166', highlight_quote='kazak contract will be ready for your review', is_interesting=True, truncate_to=1100),
    EmailCfg(
        id='EFTA00878421',
        highlight_quote="said that she felt gods presence next to her when she was in bed",
        is_interesting=True,
        recipients=[NADIA_MARCINKO],
        recipient_uncertain='https://archive.ph/Qa6vU',  # https://www.nybooks.com/articles/2026/03/26/the-devil-himself-jeffrey-epstein-enright/
    ),
    EmailCfg(id='EFTA00894079', highlight_quote='say hi to Snow White'),
    EmailCfg(id='EFTA00775255', highlight_quote='she is 20 years old but she looks younger', non_participants=[BROCK_PIERCE]),
    EmailCfg(id='EFTA00366629', highlight_quote='stunning girls to bring over drinks when your next available'),
    EmailCfg(id='EFTA02387676', highlight_quote="thats the closest i get to karyna providing me new pussy"),
    EmailCfg(id='EFTA00908180', highlight_quote='two cinderellas'),
    EmailCfg(id='EFTA01930285', highlight_quote='ukraine upheaval should provide many opportunites'),
    EmailCfg(id='EFTA00751119', highlight_quote='Valdson to teach girls how to serve'),
    EmailCfg(id='EFTA01995972', highlight_quote='you are meeting elon tmr for lunch'),
    EmailCfg(id='EFTA01998027', highlight_quote="What day/night will be the wildest party on your island?", is_interesting=True),
    EmailCfg(id='EFTA01930501', highlight_quote="Your littlest girl was a little naughty"),
    EmailCfg(id='EFTA01926961', author=SVETLANA_POZHIDAEVA, highlight_quote="If you take too many more trips to Russia, I'm going to assume that you're FSB/ KGB by your parents"),
    EmailCfg(id='DropSite 2006-05-29 1329', highlight_quote='I am giving the little girl a modeling\n> lesson'),

    # Josh Fink
    EmailCfg(
        id='EFTA01868066',
        # author=SVETLANA_POZHIDAEVA,
        # author_uncertain=f"{JOSHUA_FINK} + Sent from Blackberry",
        note=f'impersonating [someone] talking to {JOSHUA_FINK}',
        duplicate_ids=['EFTA01995523'],
        is_interesting=True,
    ),
    EmailCfg(
        id='EFTA01877084',
        comment='https://x.com/HansMahncke/status/2024351830998994979',
        note=f'Epstein and {JEAN_LUC_BRUNEL} "managing" {JOSHUA_FINK}\'s relationship with {SVETLANA_POZHIDAEVA}',
        is_interesting=True,
    ),
    EmailCfg(
        id='EFTA00704924',
        note=f"stalking Josh Fink (attachment is pic of {JOSHUA_FINK} and {FILIPA_PEROVIC} who later married)",
        show_with_name=JOSHUA_FINK,
    ),
    EmailCfg(
        id='EFTA02217672',
        note=f"{DAVID_L_NEUHAUSER} is raising money for a {JOSHUA_FINK} fund, promises Epstein he won't reveal they're talking",
        is_interesting=True,
    ),
    EmailCfg(
        id='EFTA01907281',
        note=f"Epstein asks about {JOSHUA_FINK}, pretends it's because {TANCREDI_MARCHIOLO} asked *him* (Epstein asked Marchiolo)",
        show_with_name=JOSHUA_FINK,
        is_interesting=True,
    ),

    # Descriptions
    EmailCfg(
        id='EFTA02627361',
        note=f'Epstein sending {YULIA_DOROKHINA} {QUESTION_MARKS} to {JIDE_ZEITLIN}',
        recipients=[JIDE_ZEITLIN],
        recipient_uncertain=True,
        show_with_name=YULIA_DOROKHINA,
    ),
    EmailCfg(
        id='EFTA02434682',
        note=f"Epstein and {JEAN_LUC_BRUNEL} courting a finance bro with women",
        is_interesting=True,
        truncate_to=1500,
    ),
    EmailCfg(id='EFTA01849371', note=f"{SVETLANA_POZHIDAEVA} meets Larry Fink", truncate_to=500),
    # EmailCfg(id='EFTA01849371', note=f"{SVETLANA_POZHIDAEVA} meets Nat Rothschild", show_with_name=SVETLANA_POZHIDAEVA),
    EmailCfg(id='EFTA02605514', author=SVETLANA_POZHIDAEVA, note=f"{SVETLANA_POZHIDAEVA} meets {BILL_GATES}"),
    EmailCfg(id='EFTA01895036', note=f"Epstein invites {ELON_MUSK} to Paris", highlight_quote='svetlana will be in paris 17-20'),
    EmailCfg(id='EFTA02142175', note=f"{DARREN_INDYKE} meeting {JEAN_LUC_BRUNEL} at Karin Models office"),
    EmailCfg(id='EFTA01840103', note="Epstein apparently suggesting amphetamines (adderall)", is_interesting=10),
    EmailCfg(id='EFTA01987855', note=f'Epstein, Brunel, and {SULTAN_BIN_SULAYEM} very interested in a Liberian sex scandal'),
    EmailCfg(id='EFTA01817903', note=f'Epstein, Brunel, and {SULTAN_BIN_SULAYEM} sharing news of a Liberian sex scandal', is_interesting=10),
    EmailCfg(id='EFTA01818540', note=f'Epstein forwarding news of the Liberian sex scandal to {JES_STALEY}'),
    EmailCfg(id='EFTA01140210', note='Epstein asks about fake Instagram followers'),
    EmailCfg(id='EFTA00937981', note=f"Epstein looking for {JOSHUA_FINK} (again)"),
    EmailCfg(id='EFTA00668344', note=f"Epstein looking for {JOSHUA_FINK} (again)"),
    EmailCfg(id='EFTA01856467', note=f"Epstein looking for {JOSHUA_FINK} (again)", truncate_to=NO_TRUNCATE),
    EmailCfg(id='EFTA01854125', note=f'Epstein says to {HENRY_JARECKI} "you torture, and mistreat each" girl'),
    EmailCfg(id='EFTA01936438', note=f'Epstein\'s standard password might be "MVEMJSNUP"', is_interesting=10),
    EmailCfg(id='EFTA02559808', note=f"{EVA_DUBIN} delivering amphetamines (adderall) to Epstein", is_interesting=10, truncate_to=200),
    EmailCfg(id='EFTA01953412', note=f'{EVA_DUBIN} suggests Epstein endow "The Epstein Floor For Women" at Mt. Sinai'),
    EmailCfg(id='EFTA00658028', note=f'{FAITH_KATES} trying to get a modeling contract for "Regina"?'),
    EmailCfg(id='EFTA00855566', note=f"introducing {SANITA}", truncate_to=NO_TRUNCATE),
    EmailCfg(id='EFTA01145923', note=f"{JEAN_LUC_BRUNEL} modeling contract for {MC2_MODEL_MGMT}"),
    EmailCfg(id='EFTA01660651', note='list of Trump accusers', is_interesting=True),
    EmailCfg(id='EFTA02609062', note=f"{MASHA_DROKOVA} is assembling a team", is_interesting=True),
    EmailCfg(id='EFTA02557291', note='possibly recruiting girls'),
    EmailCfg(
        id='EFTA01811230',
        note=f'"Renat" is probably {RENATA_BOLOTOVA}, {REDACTED} might be "Irina" or {SVETLANA_POZHIDAEVA}',
        duplicate_ids=['EFTA00732294'],
    ),
    EmailCfg(id='EFTA00927927', note=f"{RENATA_BOLOTOVA} possibly talking about recruiting girls"),
    EmailCfg(id='EFTA01816788', note=f"Russian girl for {PRINCE_ANDREW}"),
    EmailCfg(id='EFTA00876739', note=f'scouting girls in Ukraine {QUESTION_MARKS}'),
    EmailCfg(id='EFTA00901905', note="smoking in the house oh no!"),
    EmailCfg(id='EFTA00914505', note=f'{SVETLANA_POZHIDAEVA} and {JOSHUA_FINK}...', duplicate_ids=['EFTA01999588']),
    EmailCfg(id='EFTA01798715', note=f"taking an interest in {JOSHUA_FINK}'s investments in Eritrea in between hair gigs"),
    EmailCfg(id='EFTA00572157', note='translation: "Congratulations on the new addition!!! What did you name the little miracle?" (seems to be about a cat)'),
    EmailCfg(id='EFTA00631762', note="visa problems for Epstein's South African friend"),
    EmailCfg(id='EFTA00376832', show_with_name=MARIANA_IDZKOWSKA),
    EmailCfg(id='EFTA00743526', truncate_to=NO_TRUNCATE),
    EmailCfg(id='EFTA02664956', truncate_to=NO_TRUNCATE),
    EmailCfg(id='EFTA00689698', truncate_to=NO_TRUNCATE),
    EmailCfg(id='EFTA02702141', truncate_to=NO_TRUNCATE, comment=ELON_MUSK),
    EmailCfg(id='EFTA00819508', truncate_to=NO_TRUNCATE, comment=JULIA_SANTOS),
    EmailCfg(id='EFTA00460312', truncate_to=NO_TRUNCATE, comment=JULIA_SANTOS),
    EmailCfg(id='EFTA00878255', comment=BORIS_NIKOLIC),
    EmailCfg(id='EFTA02551185', comment=BORIS_NIKOLIC),
    EmailCfg(id='EFTA02538269', comment=BORIS_NIKOLIC, duplicate_ids=['EFTA01866636']),
    EmailCfg(id='EFTA00736518', comment=JES_STALEY + ' beauty and the beast'),
    EmailCfg(id='EFTA02344845', comment=JULIA_SANTOS),
    EmailCfg(id='EFTA02491007', comment=JULIA_SANTOS),
    EmailCfg(id='EFTA00633284', comment=JULIA_SANTOS),
    EmailCfg(id='EFTA00664521', comment=JULIA_SANTOS),
    EmailCfg(id='EFTA00997666', comment=JULIA_SANTOS),
    EmailCfg(id='EFTA00964428', comment=JULIA_SANTOS + ' Valdson'),
    EmailCfg(id='EFTA02398135', comment=JULIA_SANTOS + ' "Sochi"'),
    EmailCfg(id='EFTA00766770'),
]



import logging
from dataclasses import dataclass, field
from datetime import datetime, date
from typing import Sequence, Self

from rich.align import Align
from rich.console import Group, RenderableType
from rich.padding import Padding
from rich.panel import Panel
from rich.table import Table
from rich.text import Text

from epstein_files.documents.communication import Communication
from epstein_files.documents.document import Document
from epstein_files.documents.documents.categories import Uninteresting
from epstein_files.documents.documents.doc_types_mixin import DocTypesMixin
from epstein_files.documents.email import BCC_LISTS, TRUNCATE_EMAILS_BY, MAILING_LISTS, Email
from epstein_files.documents.emails.emailers import ENTITIES_DICT, cleanup_str, get_entity
from epstein_files.documents.messenger_log import MessengerLog
from epstein_files.documents.other_file import OtherFile
from epstein_files.output.highlight_config import (HIGHLIGHTED_NAMES, QUESTION_MARKS_TXT, get_highlight_group_for_name,
     get_style_for_name, styled_category, styled_name)
from epstein_files.output.highlighted_names import HighlightedNames, HighlightPatterns, ManualHighlight
from epstein_files.output.layout_elements.file_display import FileDisplay
from epstein_files.output.rich import GREY_NUMBERS, TABLE_TITLE_STYLE, build_table, console, print_special_note
from epstein_files.output.site.internal_links import TO_FROM
from epstein_files.people.entity import Entity
from epstein_files.people.interesting_people import SPECIAL_NOTES
from epstein_files.util.constant.strings import *
from epstein_files.util.constant.urls import *
from epstein_files.util.constants import *
from epstein_files.util.env import args, site_config
from epstein_files.util.helpers.data_helpers import days_between, flatten, uniquify, without_falsey
from epstein_files.util.external_link import join_texts
from epstein_files.util.logging_entity import LoggingEntity

ALT_INFO_STYLE = 'medium_purple4'
CC = 'cc:'
MAX_NAME_COL_WIDTH = 24
MAX_INFO_COL_WIDTH = len("Epstein's Russian assistant who was recommended for a visa by Sergei Belyakov")
MIN_AUTHOR_PANEL_WIDTH = 80
EMAILER_INFO_TITLE = 'Email Conversations Will Appear'
SOLE_CC_STYLE = 'wheat4 dim'
SOLE_CC_NO_CONTACT_STYLE = 'plum4 dim'
UNINTERESTING_CC_INFO = "cc: or bcc: recipient only"
UNINTERESTING_CC_INFO_NO_CONTACT = f"{UNINTERESTING_CC_INFO}, no direct contact with Epstein"

# TODO: get rid of this
PEOPLE_BIOS = {
    contact.name: contact.bio_txt
    for highlighted_group in HIGHLIGHTED_NAMES
    for contact in highlighted_group.entities
    if contact.has_bio
}

# Preconfigured special cases
TABLE_TXTS = {
    None: Text('(emails whose author or recipient could not be determined)', style=ALT_INFO_STYLE),
    JEFFREY_EPSTEIN: Text('(emails sent by Epstein to himself are here)', style=ALT_INFO_STYLE),
    Uninteresting.JUNK: Text(f"({Uninteresting.JUNK} mail)", style='bright_black dim'),
}


@dataclass
class Person(DocTypesMixin, LoggingEntity):
    """
    Collects all known info and files connected to someone who is the author or recipient
    of at least one Epstein File with methods to work with that collection as a whole.
    """
    name: Name = None
    is_interesting: bool | None = None
    entity: Entity = field(init=False)  # TODO: rename 'entity'
    is_configured_entity: bool = False  # True if Entity is configured, False if it's derived for e.g. uninteresting CCs
    _searched_for_highlight_group: bool = False

    def __post_init__(self):
        self._documents = Document.sort_by_timestamp(self.documents)
        self._populate_entity()

    @property
    def category(self) -> str:
        """Categories are configured with the `HighlightedGroups`."""
        if self.entity.category and self.entity.category == self.name:
            self._warn(f"category same as name")

        return self.entity.category

    @property
    def category_txt(self) -> Text | None:
        if self.name is None:
            return None
        elif self.category:
            return styled_category(self.category)
        elif self.is_a_mystery or self.is_interesting is False:
            return QUESTION_MARKS_TXT

    @property
    def counterparties(self) -> list[Name]:
        """All text and email counterparties for this person."""
        all_counterparties = flatten([c.participants for c in self.communications])
        return sort_names([c for c in all_counterparties if c != self.name])

    @property
    def full_info_panel(self) -> Padding:
        """Return a `Group` with the name of an emailer and a few tidbits of information about them below."""
        elements: list[RenderableType] = [self.email_info_panel()]

        if self.header_panel_bio_txt:
            elements.append(self.header_panel_bio_txt)

        return Padding(Group(*elements), (2, 0, 1, 0))

    @property
    def email_conversation_length_in_days(self) -> int:
        """Number of days between earliest and most recent emails."""
        return days_between(self.emails[0].timestamp, self.emails[-1].timestamp)

    @property
    def emails_by(self) -> list[Email]:
        """Emails written by this person."""
        return [e for e in self.emails if self.name == e.author]

    @property
    def emails_to(self) -> list[Email]:
        """Emails sent to this person."""
        return [
            e for e in self.emails
            if self.name in e.recipients or (self.name is None and len(e.recipients) == 0)
        ]

    @property
    def has_any_epstein_emails(self) -> bool:
        """True if any emails sent to or received from Jeffrey Epstein."""
        contacts = [e.author for e in self.emails] + flatten([e.recipients for e in self.emails])
        return JEFFREY_EPSTEIN in contacts

    @property
    def header_panel_bio_txt(self) -> Text | None:
        """The text that appears as the bio line in the header panel for a person's emails."""
        if self.is_configured_entity and self.entity.info_with_category:
            return Text(f"({self.entity.info_with_category})", justify='center', style=f"{self._email_info_style} italic")

    @property
    def is_a_mystery(self) -> bool:
        """Return True if this is someone we theroetically could know more about."""
        return not (self.entity.is_email_address or self.entity.info or self.is_interesting is False)

    @property
    def name_str(self) -> str:
        return self.name or UNKNOWN

    @property
    def name_txt(self) -> Text:
        return styled_name(self.name_str)

    @property
    def other_files_shown_with_emails(self) -> list[OtherFile]:
        """OtherFile objects that should be displayed in the emails section(s)."""
        return [f for f in self.other_files if self.name == f._config.show_with_name]

    @property  # TODO: unused?
    def should_always_truncate(self) -> bool:
        """True if we want to truncate all emails to/from this user."""
        return self.name in TRUNCATE_EMAILS_BY or self.is_interesting is False

    @property
    def sole_cc_author(self) -> str | None:
        """Return a name if this person sent 0 emails (total) and received CCs from only that one name."""
        email_authors = uniquify([e.author for e in self.emails_to])

        if self.num_emails == 1 and len(email_authors) > 0:
            logger.info(f"sole author of email to '{self.name}' is '{email_authors[0]}'")
        else:
            logger.info(f"'{self.name}' email_authors '{email_authors[0]}'")

        if len(self.unique_emails_by) > 0:
            return None

        if len(email_authors) == 1:
            return email_authors[0]

    @property
    def sort_key(self) -> list[int | str]:
        """Key used to sort `Person` objects by the number of emails sent/received + interestingness."""
        counts = [
            self.num_emails,
            -1 * int(UNINTERESTING_CC_INFO_NO_CONTACT in self.entity.info),
            -1 * int(UNINTERESTING_CC_INFO in self.entity.info),
            int(self.has_any_epstein_emails),
        ]

        counts = [-1 * count for count in counts]

        if args.sort_alphabetical:
            return [self.name_str] + counts
        else:
            return counts + [self.name_str]

    @property
    def table_txt(self) -> Text | None:
        """Text that appears next to this name in tables of emailers."""
        self._populate_entity()
        for table_txt_key in [self.name, self.entity.category]:
            if table_txt_key in TABLE_TXTS:
                return TABLE_TXTS[table_txt_key]     # Return preconfigured in some cases

        # TODO: this sucks
        style = self.entity._style.dim if self.entity.info.startswith('(') else self.entity._style.not_bold
        return Text(self.entity.info, style)

    @property
    def unique_emails_by(self) -> list[Email]:
        return Document.without_dupes(self.emails_by)

    @property
    def unique_emails_to(self) -> list[Email]:
        return Document.without_dupes(self.emails_to)

    @property
    def _email_info_style(self) -> str:
        return 'white' if (not self.style() or self.style() == DEFAULT) else self.style()

    @property
    def _identifier(self) -> str:
        """Required `LoggingEntity` mixin."""
        return self.name_str

    @property
    def _printable_emails(self):
        """For Epstein we only want to print emails he sent to himself."""
        if self.name == JEFFREY_EPSTEIN:
            return [e for e in self.emails if e.is_note_to_self()]
        else:
            return self.emails

    @property
    def _unique_printable_emails(self):
        return Document.without_dupes(self._printable_emails)

    def email_info_panel(self) -> Panel:
        """Just the person's name on the colored background with email counts."""
        if self.name == JEFFREY_EPSTEIN:
            email_count = len(self._printable_emails)
            title_suffix = f"sent by {JEFFREY_EPSTEIN} to himself"
        else:
            email_count = self.num_emails
            num_days = self.email_conversation_length_in_days
            title_suffix = f"{TO_FROM} {self.name_str} starting {self.earliest_email_date} covering {num_days:,} days"

        title = f"Found {email_count} emails {title_suffix}"
        width = max(MIN_AUTHOR_PANEL_WIDTH, len(title) + 4, len(self.entity.info_with_category) + 8)
        panel_style = f"black on {self._email_info_style} bold"
        return Panel(Text(title, justify='center'), width=width, style=panel_style)

    def print_emails(self, printer: 'DocPrinter') -> list[Email]:
        """
        Print complete emails to or from a particular 'author' along with any specially marked docs
        configured with `show_with_name` of this user. Returns the Emails that were printed.
        """
        printer.print_centered(self.email_info_panel())

        if self.header_panel_bio_txt:
            printer.print_centered(self.header_panel_bio_txt)

        printer.line()

        if site_config.show_emailer_tables:
            printer.print_centered(self._emails_table())

        if self.category == Uninteresting.JUNK:
            logger.warning(f"Not printing junk emailer '{self.name}'")  # Junk emailers only get a table
            return self._printable_emails
        elif self.name in SPECIAL_NOTES:
            # TODO: DocPrinter doesn't render HTML for the special notes
            print_special_note(SPECIAL_NOTES[self.name])

        docs = Document.sort_by_timestamp(self._printable_emails + self.other_files_shown_with_emails)
        docs = [d.file_display(align='right') if isinstance(d, OtherFile) else d for d in docs]   # TODO this sucks

        # TODO this sucks
        for d in docs:
            if isinstance(d, FileDisplay):
                d.indent = site_config.indents.show_with

        printer.print_documents(docs, log_sfx=f"[{self.name}]")
        printer.line(2)
        return self._printable_emails  # TODO: doesn't return FileDisplay objects that may have also been printed!

    def style(self, allow_bold: bool = True) -> str:
        return get_style_for_name(self.name, allow_bold=allow_bold)

    def _emails_table(self) -> Align | Padding:
        """Build a table of this person's emails summary (timestamps, subject liness, etc)."""
        # TODO: i don't think captions render in custom HTML correctly
        caption = self.entity.epstein_sites_all_links if self.entity.is_linkable else None
        table = Email.build_emails_table(self._unique_printable_emails, self.name, caption=caption)
        padded_table = Padding(table, (0, 5, 0, 5))
        logger.debug(f"built emails table for '{self.name}' with {len(table.columns)} cols and {table.row_count} rows")
        return padded_table

    def _populate_entity(self) -> None:
        """Construct a fallback `Entity` for unconfigured names."""
        self.entity = get_entity(self.name or UNKNOWN)

        if self.entity.info:
            self.is_configured_entity = True
            return

        # Fallback to style and category from regex match on string
        if (highlight_group := get_highlight_group_for_name(self.name)):
            self.entity.style = self.entity.style or highlight_group.style

            if isinstance(highlight_group, HighlightedNames):
                self.entity.category = self.entity.category or highlight_group.category_str

        # Set Entity's info + style for uninteresting CCs
        if len(self.emails_by) == 0 and self.is_interesting is False:
            if self.entity.style:
                self.entity._debug_log(f'already has style {self.entity.style}')

            # A person who wrote no emails stil might be interesting if they received email directly from Epstein
            if (lone_sender := self.sole_cc_author):
                info = f"cc: from {lone_sender} only"
                self.entity.style = self.entity.style or SOLE_CC_STYLE
            if self.has_any_epstein_emails:
                info = UNINTERESTING_CC_INFO
                self.entity.style = self.entity.style or SOLE_CC_STYLE
            else:
                info = UNINTERESTING_CC_INFO_NO_CONTACT
                self.entity.style = self.entity.style or SOLE_CC_NO_CONTACT_STYLE

            self.entity.info = f"({info})"
        elif self.name in MAILING_LISTS:
            self.entity.info = '(mailing list)'
            self.entity.style = self.entity.style or 'pale_turquoise4 dim'
        elif self.is_a_mystery or self.entity.style:
            self.entity.info = QUESTION_MARKS
            self.entity.style = self.entity.style or 'honeydew2 bold'
            self.entity.style = self.entity._style.dim

        if not self.entity.info:
            self.entity._warn(f"no entity info determinable...")

    def __str__(self):
        return f"{self.name_str}"

    @classmethod
    def emailer_info_table(
        cls,
        people: list[Self],
        highlighted: list[Self] | None = None,
        show_epstein_total: bool = False  # TODO: this sucks
    ) -> Table:
        """
        Table of info about people's emails.

        Args:
            people (list[Person]): Everyone who sent or received an email in the files.
            highlighted (list[Person], optional): Which emailers should be highlighted (rest will be dim).
            show_epstein_total (bool, optional): Whether to show total number of emails or just notes to self for Epstein himself.
        """
        highlighted = highlighted or people
        highlighted_names = [p.name for p in highlighted]
        is_selection = len(people) != len(highlighted) or args.emailers_info
        all_emails = Person.emails_from_people(people)
        email_authors = [p for p in people if p.emails_by and p.name]
        attributed_emails = [email for email in all_emails if email.author]

        if is_selection:
            title = Text(f"{EMAILER_INFO_TITLE} for the Highlighted Names Only (", style=TABLE_TITLE_STYLE)
            title.append(THE_OTHER_PAGE_TXT).append(" has the rest)")
        else:
            title = f"{EMAILER_INFO_TITLE} in Chronological Order Based on Timestamp of First Email"

        footer = f"(identified {len(email_authors)} authors of {len(attributed_emails):,}" \
                 f" out of {len(all_emails):,} emails, {len(all_emails) - len(attributed_emails)} still unknown"

        if args.all_emails:
            footer += ')'
        else:
            num_uninteresting = len([p for p in people if p.is_interesting is False])
            footer += f". {num_uninteresting} uninteresting people not shown, check all emails page for details)"

        table = build_table(title, caption=footer)
        table.add_column('First')
        table.add_column('Name', max_width=MAX_NAME_COL_WIDTH, no_wrap=True)
        table.add_column('Category', justify='left', style='dim italic')
        table.add_column('Num', justify='right', style='white')
        table.add_column('Sent', justify='right', style='wheat4')
        table.add_column('Recv', justify='right', style='wheat4')
        table.add_column('Days', justify='right', style=TIMESTAMP_DIM)
        table.add_column('Info', style='white italic', max_width=MAX_INFO_COL_WIDTH)
        current_year = 1990
        current_year_month = current_year * 12
        grey_idx = 0

        for person in people:
            if person.is_interesting is False and not (args.emailers_info or args.all_emails):
                continue

            earliest_email_date = person.earliest_email_date
            is_on_page = False if show_epstein_total else person.name in highlighted_names
            year_months = (earliest_email_date.year * 12) + earliest_email_date.month

            # Color year rollovers more brightly
            if current_year != earliest_email_date.year:
                grey_idx = 0
            elif current_year_month != year_months:
                grey_idx = ((current_year_month - 1) % 12) + 1

            current_year_month = year_months
            current_year = earliest_email_date.year

            table.add_row(
                Text(str(earliest_email_date), style=f"grey{GREY_NUMBERS[0 if is_selection else grey_idx]}"),
                person.entity.internal_link if is_on_page and person.is_interesting is not False else person.name_txt,
                person.category_txt,
                f"{len(person.unique_emails if show_epstein_total else person._unique_printable_emails)}",
                str(len(person.unique_emails_by)) if len(person.unique_emails_by) > 0 else '',
                str(len(person.unique_emails_to)) if len(person.unique_emails_to) > 0 else '',
                f"{person.email_conversation_length_in_days}",
                person.table_txt or '',
                style='' if show_epstein_total or is_on_page else 'dim',
            )

        if args.mobile:
            new_table = build_table(table.title, copy_props_from=table)
            new_table.add_column(table.columns[1].header)
            new_table.add_column(table.columns[2].header)

            for i, row in enumerate(table.rows):
                new_table.add_row(table.columns[1]._cells[i], table.columns[2]._cells[i])

            return new_table

        return table

    @staticmethod
    def emails_from_people(people: list['Person']) -> Sequence[Email]:
        """Collect all unique emails from a list of `Person` objects."""
        return Document.uniquify(flatten([list(p.unique_emails) for p in people]))



from dataclasses import dataclass, field

from rich.errors import StyleSyntaxError
from rich.style import Style
from rich.theme import Theme

from epstein_files.util.constant.html import HTML_TERMINAL_THEME
from epstein_files.util.logging import logger

BOLD = Style(bold=True)
DIM = Style(dim=True)
NOT_BOLD = Style(bold=False)
DEFAULT_THEME = Theme()


@dataclass
class RichStyle:
    """Converts rich `Style` objects and style strings to HTML RGB codes."""

    _style: Style | str | None
    style: Style = field(init=False)

    def __post_init__(self):
        if isinstance(self._style, Style):
            self.style = self._style
        else:
            if self._style is None or self._style == 'none':
                self._style = ''
            elif self._style == 'dim':
                self._style = f"white {self._style}"  # Coerce gray if just 'dim'

            try:
                self.style = DEFAULT_THEME.styles.get(self._style, Style.parse(self._style))
            except StyleSyntaxError as e:
                logger.warning(f"Failed to parse style '{self._style}', attempting to resolve w/imports...")
                from epstein_files.output.rich import RICH_THEME
                self.style = RICH_THEME.styles[self._style]

    @property
    def background_color_hex(self) -> str:
        if self.style.reverse:
            if self.style.color:
                return self.style.color.get_truecolor(HTML_TERMINAL_THEME).hex
            else:
                return ''
        elif self.style.bgcolor:
            return self.style.bgcolor.get_truecolor(HTML_TERMINAL_THEME).hex
        else:
            return ''

    @property
    def bold(self) -> Style:
        """self.style but bold."""
        return Style.combine([self.style, BOLD])

    @property
    def dim(self) -> Style:
        return Style.combine([self.style, DIM])

    @property
    def foreground_color_hex(self) -> str:
        if self.style.reverse:
            if self.style.bgcolor:
                return self.style.bgcolor.get_truecolor(HTML_TERMINAL_THEME).hex
            else:
                return ''
        if self.style.color:
            return self.style.color.get_truecolor(HTML_TERMINAL_THEME).hex
        else:
            return ''

    @property
    def not_bold(self) -> Style:
        """self.style but not bold."""
        return Style.combine([self.style, NOT_BOLD])

    @property
    def to_css(self) -> dict[str, str]:
        """Create CSS properties for this style."""
        props = {}

        if self.background_color_hex:
            props['background-color'] = self.background_color_hex

        if self.foreground_color_hex:
            props['color'] = self.foreground_color_hex

        if self.style.bold:
            props['font-weight'] = 'bold'

        if self.style.italic:
            props['text-decoration'] = 'italic'

        return props

    def __str__(self) -> str:
        return f"{type(self).__name__}(style={self.style}, _style={self._style})"



"""
Methods that print the title page with all the links etc.
"""
# Rich reference: https://rich.readthedocs.io/en/latest/reference.html
from rich import box
from rich.align import Align
from rich.console import Console, Group, NewLine, RenderableType
from rich.markup import escape
from rich.panel import Panel
from rich.padding import Padding
from rich.table import Table
from rich.text import Text

from epstein_files.output.epstein_highlighter import highlighter
from epstein_files.output.layout_elements.demi_table import build_demi_table
from epstein_files.output.highlight_config import HIGHLIGHTED_NAMES
from epstein_files.output.rich import *
from epstein_files.output.site.internal_links import SECTION_ANCHORS
from epstein_files.output.site.site_config import MOBILE_WARNING_TXT
from epstein_files.people.names import UNKNOWN
from epstein_files.util.constant.strings import *
from epstein_files.util.constant.urls import *
from epstein_files.util.constants import HEADER_ABBREVIATIONS
from epstein_files.util.env import args, site_config
from epstein_files.util.external_link import SUBSTACK_POST_LINK_STYLE, join_texts, link_markup, link_text_obj, parenthesize
from epstein_files.util.helpers.rich_helpers import starred_header_txt, vertically_pad
from epstein_files.util.helpers.string_helper import starred_header
from epstein_files.util.logging import logger

SITE_GLOSSARY_MSG = f"The following views of the underlying selection of Epstein Files are available:"
YOU_ARE_HERE = Text('«').append('you are here', style='bold khaki1 blink').append('»')
SECTION_LINK_MSG = 'jump to a different section of this page'
SUBTITLE_WIDTH = 110
TITLE_WIDTH = 50

DATASET_MSG_STYLE = 'gray74'
DIRECTORY_STYLE = 'wheat4'
OTHER_PAGE_MSG_STYLE = 'gray78 dim'
STR_VAL_STYLE = 'cornsilk1 italic'
STR_VAL_STYLE_ALT = 'light_yellow3 italic'
SECTION_HEADER_STYLE = 'bold black on color(146)'

DATASET_MSG_LINES = [
    f"This dataset includes everything from the {HOUSE_OVERSIGHT_TRANCHE}",
    f"as well as a curated selection of the {DOJ_2026_TRANCHE}",
    f"with a particular focus on cryptocurrency",
    f"and women who speak Russian.",
]

DATASET_MSG_TXTS = [Text(msg, justify='center', style=DATASET_MSG_STYLE) for msg in DATASET_MSG_LINES]
DATASET_MSG = Align.center(join_texts(DATASET_MSG_TXTS, '\n'))

# label of the HighlightedNames objects colors with the style of that same HighlightedNames
COLOR_KEYS = [
    Text(highlight_group.label.replace('_', ' '), style=highlight_group.style)
    for highlight_group in sorted(HIGHLIGHTED_NAMES, key=lambda hg: hg.label)
]

SECTION_LINKS = [
    link_text_obj(internal_link_url(anchor), section_name, 'indian_red')
    for section_name, anchor in SECTION_ANCHORS.items()
]


def color_key() -> Padding:
    color_table = build_table('Rough Guide to Highlighted Colors', show_header=False)
    num_colors = len(COLOR_KEYS)
    row_number = 0

    for i in range(0, site_config.num_color_key_cols):
        color_table.add_column(f"color_col_{i}", justify='center')

    while (row_number * site_config.num_color_key_cols) < num_colors:
        idx = row_number * site_config.num_color_key_cols
        color_table.add_row(*COLOR_KEYS[idx:(idx + site_config.num_color_key_cols)])
        row_number += 1

    return vertically_pad(color_table)


def print_color_key() -> None:
    color_table = build_table('Rough Guide to Highlighted Colors', show_header=False)
    num_colors = len(COLOR_KEYS)
    row_number = 0

    for i in range(0, site_config.num_color_key_cols):
        color_table.add_column(f"color_col_{i}", justify='center')

    while (row_number * site_config.num_color_key_cols) < num_colors:
        idx = row_number * site_config.num_color_key_cols
        color_table.add_row(*COLOR_KEYS[idx:(idx + site_config.num_color_key_cols)])
        row_number += 1

    print_centered(vertically_pad(color_table))


def print_other_page_link(epstein_files: 'EpsteinFiles') -> None:
    if args._site == Site.CURATED_MOBILE:
        return
    elif args._site == Site.CURATED:
        txt = THE_OTHER_PAGE_TXT + Text(f' is uncurated and has all {len(epstein_files.emails):,} emails')
        txt.append(f" and {len(epstein_files.other_files):,} unclassifiable files")
    else:
        txt = THE_OTHER_PAGE_TXT + (f' displays a curated collection of emails and')
        txt.append(" unclassifiable files of particular interest")

    print_centered(parenthesize(txt), style=OTHER_PAGE_MSG_STYLE)
    chrono_emails_link = link_text_obj(Site.get_url(Site.EMAILS_CHRONOLOGICAL), 'a page', style='light_slate_grey bold')
    chrono_emails_txt = Text(f"there's also ").append(chrono_emails_link)
    chrono_emails_txt.append(' with all the emails in chronological order')
    print_centered(parenthesize(chrono_emails_txt), style=OTHER_PAGE_MSG_STYLE)


def section_header(msg: str, style: str = SECTION_HEADER_STYLE, is_centered: bool = False) -> Padding:
    """Make a padded Panel that's centered."""
    if args._site == Site.CURATED:
        console.line(2)
        print_section_links()

    panel = Panel(Text(msg, justify='center'), expand=True, padding=(1, 1), style=style)
    panel = Align.center(panel) if is_centered else panel
    return Padding(panel, site_config.section_header_padding)


def print_section_links(style: str = '') -> None:
    """Print links to the various sections within the curated page."""
    print_centered(build_demi_table(SECTION_LINK_MSG, SECTION_LINKS), style=style)


def title_page_top_elements() -> list[RenderableType]:
    return [
        *_header_elements(),
        _site_directory(),
        starred_header_txt(site_config.not_all_files_warning, num_spaces=0, num_stars=0),
        *DATASET_MSG_TXTS,
    ]


def title_page_bottom_elements(epstein_files: 'EpsteinFiles') -> list[RenderableType]:
    return [
        NewLine(),
        _abbreviations_table(),
        Text('External Links', style=TABLE_TITLE_STYLE),
        *[link for link in sorted(EXTERNAL_LINKS, key=lambda link: -len(link.__rich__()))],
        epstein_files.overview_table(),
        NewLine(),
        color_key(),
        Text.from_markup(f"(if you think there's an attribution error or can deanonymize an {UNKNOWN} contact {CRYPTADAMUS_X_LINK_MARKUP})", style='grey46'),
        Text.from_markup(f"(thanks to {link_markup('https://x.com/ImDrinknWyn', '@ImDrinknWyn', 'dodger_blue3')} + others for help attributing redacted emails)"),
        NewLine(),
    ]


def _abbreviations_table() -> Padding:
    table = build_table(title="Abbreviations Used Frequently In These Conversations", show_header=False)
    table.add_column("Abbreviation", justify="center", style='bold')

    table.add_column(
        "Translation",
        justify="center",
        min_width=site_config.abbreviations_width,
        style="white"
    )

    for k, v in HEADER_ABBREVIATIONS.items():
        table.add_row(highlighter(k), v)

    return vertically_pad(table)


def _bulleted_site_link(site: Site, link: Text) -> Text:
    """These go in the site directory panel."""
    you_are_here = ''

    if args.mobile:
        if site == Site.CHRONOLOGICAL_MOBILE and args.output_chrono:
            you_are_here = YOU_ARE_HERE
        elif site == Site.CURATED:
            you_are_here = YOU_ARE_HERE

        if site not in Site.mobile_compatible_types():
            link.stylize('dim strike')
    elif site == args._site:
        you_are_here = YOU_ARE_HERE

    return Text('➱ ').append(link).append(' ').append(you_are_here)


def _header_elements() -> list[RenderableType]:
    title = Text('', justify='center').append('The Epstein Files', style='underline bold')
    title_panel = Panel(title, box=box.DOUBLE_EDGE, expand=True, padding=(2, 2), style=TITLE_STYLE, width=TITLE_WIDTH)
    title_with_vertical_margin = vertically_pad(title_panel)

    # unwrapped, css_props = unwrap_rich(title_with_vertical_margin)
    # print(f"rich obj {unwrapped} has unwrapped title_panel css: {css_props}")
    # import pdb;pdb.set_trace()

    elements = [
        MOBILE_WARNING_TXT,
        title_with_vertical_margin,
        SUBSTACK_POST_TXT_MSGS_LINK,
    ]

    if not args.mobile:
        substack_link = SUBSTACK_POST_TXT_MSGS_LINK.url_link
        substack_link.stylize(f'{SUBSTACK_POST_LINK_STYLE} dim')
        elements.append(substack_link)

    elements.append(join_texts(CRYPTADAMUS_SOCIAL_LINKS, join=site_config.social_link_separator))
    return elements


def _site_directory() -> Panel:
    """Panel with links to the various sites generated by this code."""
    links_txts = [
        _bulleted_site_link(site, link)
        for site, link in Site.directory_links().items()
    ]

    if args.mobile:
        num_link_indent_spaces = 0
    else:
        max_link_len = max(len(link.plain) for link in links_txts)
        num_link_indent_spaces = max(2, int((len(SITE_GLOSSARY_MSG) - max_link_len) / 2)) - 2

    sites_txt = Text('').append(SITE_GLOSSARY_MSG, style=f'{DIRECTORY_STYLE} bold').append('\n\n')
    sites_txt.append(indent_txt(join_texts(links_txts, '\n'), num_link_indent_spaces))

    if args.mobile:
        sites_txt.append("\n\nNot all views are available on mobile.", DIRECTORY_STYLE)

    return Panel(
        sites_txt,
        border_style='dim',
        padding=(1, site_config.site_glossary_horizontal_padding)
    )



from dataclasses import dataclass, field
from typing import Literal

from rich.align import Align
from rich.console import Console, ConsoleOptions, RenderResult, RenderableType
from rich.padding import Padding
from rich.table import Table
from rich.text import Text
from rich.panel import Panel

from epstein_files.output.html.builder import (PANEL_BASE_PROPS, VERTICAL_MARGIN_EMS, border_css_props,
     one_row_table_html, render_to_html, text_to_list, text_to_div, margin_vertical_css)
from epstein_files.output.html.elements import CssProps, div_class, div_with_legend, div_tag
from epstein_files.output.html.rich_style import RichStyle
from epstein_files.output.html.positioned_rich import BLACK_BACKGROUND, PositionedRich, dimensions_to_margin_css
from epstein_files.util.env import site_config
from epstein_files.output.rich import indent_txt
from epstein_files.util.external_link import join_texts
from epstein_files.util.helpers.data_helpers import without_falsey

JustifyMethod = Literal['center', 'left', 'right']

BOTTOM_PADDING = 1
SUBHEADER_VERTICAL_MARGIN = 0.3

DOC_DIV_CSS_PROPS = {
    'display': 'flex',
    'flex-direction': 'column',
}


@dataclass(kw_only=True)
class BasePanel:
    border_style: str
    text: Text
    title: Text | None = None
    title_justify: JustifyMethod = 'right'

    def to_div(self, margins: list[int | float] | None = None) -> str:
        """Create an HTML <div> string for this panel."""
        div_props = {
            **self._base_div_css(margins),
            **PANEL_BASE_PROPS,
            **border_css_props(self.border_style)
        }

        # TODO: make the title 'dim'
        title = self.title.plain if self.title else ''
        html = render_to_html(self.text)
        return div_with_legend(html, title, div_props)

    def _base_div_css(self, margins: list[int | float] | None = None) -> CssProps:
        return dimensions_to_margin_css(margins or PositionedRich.zero_dimensions())

    def __rich__(self) -> Panel:
        return Panel(
            self.text,
            border_style=self.border_style,
            expand=False,
            title=self.title,
            title_align=self.title_justify,
        )


@dataclass(kw_only=True)
class ListPanel(BasePanel):
    text: list[Text]

    def to_div(self, margins: list[int | float] | None = None) -> str:
        """Create an HTML <div> string for this panel."""
        div_props = {
            **self._base_div_css(margins),
            'word-wrap': 'break-word',
        }

        html = text_to_list(self.text, class_name='no_bullets')
        return div_class(html, BLACK_BACKGROUND, div_props)

    def __rich__(self) -> Panel | Text:
        txts = join_texts(self.text, '\n')

        if self.border_style or self.title:
            return Panel(
                txts,
                border_style=self.border_style,
                expand=False,
                title=self.title,
                title_align=self.title_justify,
            )
        else:
            return indent_txt(txts, site_config.indents.supressed_msg)


@dataclass(kw_only=True)
class FileDisplay:
    """Allows for proper right vs. left justify of a Document display."""
    background_color: str = ''
    body_panel: BasePanel | Table
    document: 'Document'
    file_info: BasePanel | None = None
    indent: int | float = 0  # TODO: this should be a property of the BasePanel
    justify: JustifyMethod | None = None
    margin_bottom: str = VERTICAL_MARGIN_EMS  # Margin below the entire agglomeration of elements, not just the body
    subheaders: list[Text] = field(default_factory=list)

    @property
    def body_margin(self) -> list[int | float]:
        """Margin for the body. Top/bottom always zero currrently."""
        padding = PositionedRich.zero_dimensions()

        # Set subtle indent
        if self.justify == 'right':
            padding[1] = self.indent
        else:
            padding[3] = self.indent

        return padding

    @property
    def body_margin_horizontal(self) -> list[int | float]:
        """Just left and right margin, vertical margins are zeroed out."""
        margin_dimensions = self.body_margin
        margin_dimensions[0] = margin_dimensions[2] = 0
        return margin_dimensions

    @property
    def horizontal_body_margin_css(self) -> dict[str, str]:
        return dimensions_to_margin_css(self.body_margin_horizontal)

    @property
    def is_table(self) -> bool:
        return isinstance(self.body_panel, Table)

    @property
    def justified_subheaders(self) -> list[Text]:
        """Set justify `Text` to get right justification of text in right aligned elements."""
        subheaders = [txt.copy() for txt in self.subheaders]

        if self.justify:
            for txt in subheaders:
                txt.justify = self.justify

        return subheaders

    @property
    def subheader_div(self) -> str:
        if not self.subheaders:
            return ''

        css_props = {
            **self.horizontal_body_margin_css,
            **margin_vertical_css(SUBHEADER_VERTICAL_MARGIN),
        }

        return text_to_div(Text('\n').join(self.subheaders), css_props)

    def to_html(self) -> str:
        css_props = {
            **DOC_DIV_CSS_PROPS,
            'margin-bottom': self.margin_bottom,
        }

        if self.background_color:
            css_props.update(RichStyle(f"on {self.background_color}").to_css)

        # Add more vertical margin before/after text messages  # TODO: this shouold be configured
        if isinstance(self.body_panel, ListPanel):
            css_props.update(margin_vertical_css(2))

        if self.is_table:
            body_html = one_row_table_html(self.body_panel, self.horizontal_body_margin_css)
        else:
            body_html = self.body_panel.to_div(self.body_margin_horizontal)

        elements = [
            self.file_info.to_div() if self.file_info else None,
            self.subheader_div,
            body_html,
        ]

        return div_tag(without_falsey(elements), css_props)

    def _align(self, element: RenderableType) -> RenderableType:
        return Align(element, self.justify) if self.justify else element

    def __rich_console__(self, console: Console, options: ConsoleOptions) -> RenderResult:
        """Default `Document` renderer (Email and MessengerLog override this)."""
        # Set justify on the Text in the body panel
        if self.justify and not self.is_table and isinstance(self.body_panel.text, Text):
            self.body_panel.text.justify = self.justify

        indented_elemeents = [*self.justified_subheaders, self.body_panel]
        indented_elemeents = [Padding(e, self.body_margin) for e in indented_elemeents]
        indented_elemeents[-1].bottom = BOTTOM_PADDING
        elements = ([self.file_info] if self.file_info else []) + indented_elemeents

        for element in elements:
            yield self._align(element)



import logging
import re
from collections import Counter
from contextlib import contextmanager
from dataclasses import asdict, dataclass, field
from datetime import datetime
from email import policy
from email.parser import BytesParser
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import ClassVar, Generator, Mapping, Self, Sequence, TypeVar

from rich.align import Align
from inflection import underscore
from rich.console import Console, ConsoleOptions, Group, JustifyMethod, RenderResult
from rich.markup import escape
from rich.padding import Padding
from rich.panel import Panel
from rich.text import Text
from rich.table import Table

from epstein_files.documents.config.doc_cfg import DEFAULT_TRUNCATE_TO, DUPE_TYPE_STRS, NO_TRUNCATE, DebugDict, DocCfg, Metadata
from epstein_files.documents.config.email_cfg import EmailCfg
from epstein_files.documents.documents.file_info import FileInfo
from epstein_files.documents.documents.search_result import MatchedLine
from epstein_files.documents.emails.constants import DOJ_EMAIL_OCR_REPAIRS, FALLBACK_TIMESTAMP
from epstein_files.documents.emails.emailers import get_entities
from epstein_files.documents.emails.email_header import DETECT_EMAIL_REGEX
from epstein_files.output.highlight_config import HIGHLIGHTED_ENTITIES, get_style_for_category, get_style_for_name
from epstein_files.output.html.builder import VERTICAL_MARGIN_EMS
from epstein_files.output.layout_elements.file_display import BasePanel, FileDisplay
from epstein_files.output.rich import (INFO_STYLE, NA_TXT, SYMBOL_STYLE, add_cols_to_table, build_table, console,
     styled_key_value, prefix_with, snip_msg_txt, styled_dict)
from epstein_files.output.site.sites import EXTRACTS_BASE_URL
from epstein_files.people.entity import Entity, EntityScanArg
from epstein_files.people.interesting_people import PERSONS_OF_INTEREST, UNINTERESTING_AUTHORS
from epstein_files.people.names import Name
from epstein_files.util.constant.strings import *
from epstein_files.util.constants import CONFIGS_BY_ID
from epstein_files.util.env import args, site_config, temporary_args
from epstein_files.util.external_link import link_text_obj
from epstein_files.util.helpers.data_helpers import (CharRange, coerce_utc, coerce_utc_strict, date_str, patternize, prefix_keys,
     listify, uniquify, uniq_sorted, without_falsey)
from epstein_files.util.helpers.file_helper import coerce_file_path, file_size_str, file_size_to_str
from epstein_files.util.helpers.rich_helpers import TextVar
from epstein_files.util.helpers.string_helper import collapse_newlines, doublespace_lines, join_truthy, quote, timestamp_without_zero_hour
from epstein_files.util.logging import DOC_TYPE_STYLES, FILENAME_STYLE, logger
from epstein_files.util.logging_entity import LoggingEntity

CLOSE_PROPERTIES_CHAR = ']'
HOUSE_OVERSIGHT = HOUSE_OVERSIGHT_PREFIX.replace('_', ' ').strip()
MIN_DOCUMENT_ID = 10477

FILENAME_MATCH_STYLES = [
    'dark_green',
    'green',
    'spring_green4',
]

OCR_REPAIRS = {
    re.compile(r'\.corn\b'): '.com',
    re.compile('ln(adequate|dyke)'): r'In\1',
    'Nil Priell': 'Nili Priell',
}

SUMMARY_TABLE_COLS: list[str | dict] = [
    'Count',
    {'name': 'Has Author', 'style': 'honeydew2'},
    {'name': 'No Author', 'style': 'wheat4'},
    {'name': 'Uncertain Author', 'style': 'royal_blue1 dim'},
    {'name': 'Size', 'justify': 'right', 'style': 'dim'},
]

DEBUG_PROPS = [
    'is_interesting',
    'num_lines',
    'timestamp',
]

DEBUG_PROPS_TRUTHY_ONLY = [
    AUTHOR,
    'category',
    'is_empty',
]

METADATA_FIELDS = [
    AUTHOR,
    'file_id',
    'filename',
    'num_lines',
    'timestamp',
]

T = TypeVar('T', bound=str | Text)


@dataclass
class Document(LoggingEntity):
    """
    Base class for all Epstein Files documents.

    Attributes:
        file_path (Path): Local path to file
        extracted_author (Name): who created the text in this file, extracted from the text (AKA "not configured")
        extracted_timestamp (datetime, optional): When the file was originally created, extracted from the text
        file_info (FileInfo): Manages things having to do with the underlying file (paths, URLs, etc.)
        lines (list[str]): Number of lines in the file after all the cleanup
        text (str): Contents of the file
    """
    file_path: Path

    # Derived at instantiation time
    extracted_author: Name = None
    extracted_timestamp: datetime | None = None
    file_info: FileInfo = field(init=False)
    lines: list[str] = field(default_factory=list)
    text: str = ''

    # Class constants, overloaded in some subclasses
    MAX_TIMESTAMP: ClassVar[datetime] = coerce_utc_strict(datetime(2026, 1, 29))  # Cutoff for extract_timestamp()
    STRIP_WHITESPACE: ClassVar[bool] = True                                       # Should strip whitespace (overridden in JsonFile)
    _INCLUDE_DESCRIPTION_IN_SUMMARY_PANEL: ClassVar[bool] = False                 # For logging only

    def __post_init__(self):
        self.file_info = FileInfo(self.file_path)

        if self.file_info.has_file and not self.file_path.exists():
            raise FileNotFoundError(f"File '{self.file_path}' does not exist!")

        self._set_text(text=self.text or self._load_file())
        self._repair()
        self.extracted_author = None if self.author else self.extract_author()
        self.extracted_timestamp = None if self.timestamp else coerce_utc(self.extract_timestamp())

    @classmethod
    def from_file_id(cls, file_id: str | int) -> Self:
        """Alternate constructor that finds the file path automatically and builds a `Document`."""
        return cls(coerce_file_path(file_id))

    @property
    def author(self) -> Name:
        return self._config.author or self.extracted_author

    @property
    def author_style(self) -> str:
        return get_style_for_name(self.author)

    @property
    def border_style(self) -> str:
        """Should be overloaded in subclasses."""
        return 'white'

    @property
    def category(self) -> str:
        return self._config.category or self.default_category()

    @property
    def category_style(self) -> str:
        return get_style_for_category(self.category) or ''

    @property
    def config(self) -> DocCfg | None:
        """Get the configured `DocCfg` object for this file id (if any)."""
        return CONFIGS_BY_ID.get(self.file_id)

    # TODO: swap this in for config() when it's safe to do so
    @property
    def _config(self) -> DocCfg:
        """Like `self.config` but falls back to return empty `DocCfg` object."""
        return self.config or self.dummy_cfg()

    @property
    def colored_external_links(self) -> Text:
        """Collection of links to official and unofficial for this file concatenated into one `Text` object."""
        return self.file_info.build_external_links(with_alt_links=True)

    @property
    def date_str(self) -> str | None:
        return date_str(self.timestamp)

    @property
    def char_range_to_display(self) -> CharRange | None:
        """Index of first and last characters to show when printing this document."""
        if args.whole_file or self._config.truncate_at == NO_TRUNCATE:
            return None
        elif args.truncate:
            return (0, args.truncate)
        elif self._config.char_range:
            return self._config.char_range

    @property
    def display_text(self) -> str:
        """Config overrides what text should be displayed."""
        return collapse_newlines(self._config.display_text or self.text)

    @property
    def duplicate_file_txt(self) -> Text | None:
        """If the file is a dupe make a nice message to explain what file it's a duplicate of."""
        if self.is_duplicate:
            # TODO: this link is incorrect for DOJ files
            link = link_text_obj(FileInfo.url_for_id(self.file_info.url_slug), self.file_id, 'royal_blue1')
            return self._skipped_file_txt(Text(f"{DUPE_TYPE_STRS[self.config.dupe_type]} ").append(link))

    @property
    def empty_file_txt(self) -> Text | None:
        """Overridden in DojFile."""
        pass

    @property
    def external_link_markup(self) -> str:
        """Rich markup string with link to source document."""
        return self.file_info.external_link_markup(get_style_for_name(self.author) if self.author else '')

    @property
    def external_link_txt(self) -> Text:
        return Text.from_markup(self.external_link_markup)

    @property
    def file_id(self) -> str:
        return self.file_info.file_id

    @property
    def file_id_panel(self) -> BasePanel:
        """The header panel printed before the body and subheaders with links and file ID etc."""
        return BasePanel(border_style=self.border_style, text=self.colored_external_links)

    @property
    def filename(self) -> str:
        return self.file_info.filename

    @property
    def html_margin_bottom(self) -> str:
        """Overloaded in `Email` for case of emails with attachments."""
        return VERTICAL_MARGIN_EMS

    @property
    def is_duplicate(self) -> bool:
        return bool(self._config.duplicate_of_id)

    @property
    def is_email(self) -> bool:
        """True if the text looks like it's probably an email."""
        search_area = self.repair_ocr_text(DOJ_EMAIL_OCR_REPAIRS, self.text[0:5000])
        return isinstance(self.config, EmailCfg) or bool(DETECT_EMAIL_REGEX.match(search_area) and self.config is None)

    @property
    def is_empty(self) -> bool:
        return len(self.text.strip()) == 0

    @property
    def is_interesting(self) -> bool | None:
        """
        If `self.config.is_of_interest` returns a boolean (not `None`) return that, otherwise:

            1. this doc is attached to an `Email`:                             false     (the interestingness of the `Email` decides)
            2. `self.author` in `UNINTERESTING_AUTHORS`:                       false
            3. `HOUSE_OVERSIGHT_XXXXX` files with no author or `DocCfg`:      **TRUE**
            4. the configured `show_with_name` in PERSONS_OF_INTEREST:        **TRUE**

        `None` is returned if no decision is made so subclasses can append additional conditions.
        """
        if (is_of_interest := self._config.is_of_interest) is not None:
            return is_of_interest
        elif self._config.attached_to_email_id:
            return False
        elif self.author in UNINTERESTING_AUTHORS:
            return False
        elif self.file_info.is_house_oversight_file and not (self.author or self.config):
            return True
        elif self._config.show_with_name in PERSONS_OF_INTEREST:
            return True
        else:
            return None

    @property
    def length(self) -> int:
        return len(self.text)

    @property
    def metadata(self) -> Metadata:
        metadata = self.config.metadata if self.config else {}
        metadata.update({k: getattr(self, k) for k in METADATA_FIELDS if getattr(self, k) is not None})
        metadata['file_size'] = self.file_info.file_size
        metadata['type'] = self._class_name

        if self.file_info.is_local_extract_file:
            metadata['extracted_file'] = {
                'explanation': 'manually extracted from one of the other files',
                'extracted_from': self.file_info.url_slug + '.txt',
                'url': f"{EXTRACTS_BASE_URL}/{self.filename}",
            }

        return metadata

    @property
    def num_lines(self) -> int:
        return len(self.lines)

    @property
    def panel_title_timestamp(self) -> str:
        """Time or date shown in the `title` of the enclosing `Panel` when printing this `Document`."""
        if self.timestamp is None or self.timestamp == FALLBACK_TIMESTAMP:
            return ''
        elif self._config.timestamp:
            prefix = 'approximate' if self._config.date_uncertain else ''
        else:
            prefix = 'inferred'

        if all(unit == 0 for unit in [self.timestamp.hour, self.timestamp.minute, self.timestamp.second]):
            date_or_time = 'date'
        else:
            date_or_time = 'timestamp'

        return join_truthy(prefix, f"{date_or_time}: {timestamp_without_zero_hour(self.timestamp)}")

    @property
    def entities(self) -> list[Entity]:
        """Every configured name that is either the author or appears in the text or description."""
        return self.entity_scan()

    @property
    def entity_names(self) -> list[str]:
        return [e.name for e in self.entities]

    @property
    def prettified_txt(self) -> Text:
        """Returns the string we want to print as the body of the document."""
        display_text = doublespace_lines(self.display_text)
        # TODO: do something better to give replacement_text have different style
        style = INFO_STYLE if self._config.has_full_ocr_text_replacement else ''

        # char range slice of Text late in the game here preserves Text highlighting at boundaries
        char_range = self.char_range_to_display or (0, len(display_text))
        selected_txt = self._config.text_highlighter(Text(display_text, style))[char_range[0]:char_range[1]]
        pretty_txt = self._intro_txt(char_range[0]).append(selected_txt)

        # For debugging/choosing truncation points only
        if args.char_nums:
            pretty_txt = self._inject_line_numbers(pretty_txt, args.char_nums)

        if (footer_txt := self.trimmed_chars_txt(char_range[1])):
            pretty_txt.append('...\n\n').append(footer_txt)

        return pretty_txt

    @property
    def subheader_info(self) -> Text | None:
        """Secondary info about this file (description, recipients, etc). Overload in subclasses."""
        return None

    @property
    def subheaders(self) -> list[Text]:
        """0 to 2 sentences containing the `subheader_info` and any configured `note` text."""
        return without_falsey([self.subheader_info, self._config.note_txt])

    @property
    def suppressed_txt(self) -> Text | None:
        """`Text` object to print if this documents is suppressed for various reasons."""
        return self.uninteresting_txt or self.duplicate_file_txt or self.empty_file_txt

    @property
    def timestamp(self) -> datetime | None:
        """Configuration timestamp takes precedence over extracted / derived timestamp."""
        return self._config.timestamp or self.extracted_timestamp

    @property
    def timestamp_sort_key(self) -> tuple[datetime, str, int]:
        """Sort by timestamp, file_id, then whether or not it's a duplicate file."""
        if self._config.duplicate_of_id:
            sort_id = self._config.duplicate_of_id
            dupe_idx = 1
        else:
            sort_id = self.file_id
            dupe_idx = 0

        return (self.timestamp or FALLBACK_TIMESTAMP, sort_id, dupe_idx)

    @property
    def uninteresting_txt(self) -> Text | None:
        """Text to print for uninteresting files."""
        if self._config.attached_to_email_id:
            self._warn(f"returning email attachment uninteresting_txt, probably shouldn't be called...")
            return self._skipped_file_txt(f"attached to {self._config.attached_to_email_id}")
        elif args._suppress_uninteresting and self._config.is_interesting is False:
            return self._skipped_file_txt("uninteresting")
        else:
            return None

    @property
    def _class_style(self) -> str:
        return DOC_TYPE_STYLES[self._class_name]

    @property
    def _debug_prefix(self) -> str:
        return underscore(self._class_name).lower()

    @property
    def _identifier(self) -> str:
        """Required `LoggingEntity` abstract method."""
        return self.file_id

    @property
    def _log_prefix(self) -> str:
        """`Overload default LoggingEntity` method."""
        return f"{self.file_id} {self._class_name}"

    @property
    def _summary(self) -> Text:
        """Summary of this file for logging. Subclasses should extend with a method that closes the open '['."""
        txt = Text('').append(self._class_name, style=self._class_style)
        txt.append(f" {self.file_id}", style=FILENAME_STYLE)

        if self.timestamp:
            timestamp_str = timestamp_without_zero_hour(self.timestamp)
            txt.append(' (', style=SYMBOL_STYLE)
            txt.append(f"{timestamp_str}", style=TIMESTAMP_DIM).append(')', style=SYMBOL_STYLE)

        txt.append(' [').append(styled_key_value('size', Text(str(self.length), style='aquamarine1')))
        txt.append(", ").append(styled_key_value('lines', self.num_lines))

        if self._config.duplicate_of_id:
            txt.append(", ").append(styled_key_value('dupe_of', Text(self._config.duplicate_of_id, style='cyan dim')))

        return txt

    @property
    def _summary_panel(self) -> Panel:
        """Panelized `subheaders`. Used in search results not in production HTML."""
        sentences = [self._summary]

        if self._INCLUDE_DESCRIPTION_IN_SUMMARY_PANEL:
            sentences += [Text('', style='italic').append(h) for h in self.subheaders]

        return Panel(Group(*sentences), border_style=self._class_style, expand=False)

    def entity_scan(self, exclude: EntityScanArg = None, include: EntityScanArg = None) -> list[Entity]:
        """
        Find people and companies associated with this document.
        `config.entity_names` overrides everything.
        `config.is_valid_for_name_scan` means don't scan text.

        Args:
            `exclude` (EntityScanArg, optional): optimization to avoid expensive rescans if bio was already printed.
            `include` (EntityScanArg, optional): additional names to include (used by subclass overrides)
        """
        entities = get_entities([self.author] + self._config.entity_names + self._coerce_entities(include))
        excluded_names = Entity.coerce_entity_names(exclude) + self._config.non_participants

        if self._config.entity_names or not self._config.is_valid_for_name_scan:
            return [e for e in entities if e.name not in excluded_names]

        text_to_scan = join_truthy(self._config.note, self.display_text, '\n')

        entities += [
            c for c in HIGHLIGHTED_ENTITIES
            if c.name not in excluded_names and c.highlight_regex.search(text_to_scan)  # excluded 1st to avoid expensive scans
        ]

        return self._coerce_entities(entities)

    def excerpt_text(self, char_range: CharRange | None = None, text: str = '', style = '') -> Text:
        """Create an excerpt of `text`, add appropriate header/footer if truncated, and highlight it."""
        char_range = char_range or (0, len(text))
        text = doublespace_lines(text or self.text)        # TODO: not ideal place for doublespace call
        excerpt_txt = self._intro_txt(char_range[0])
        excerpt_txt.append(Text(text, style)[char_range[0]:char_range[1]])  # array slice of `Text` obj preserves style

        if args.char_nums:    # For debugging/choosing truncation points
            excerpt_txt = self._inject_line_numbers(excerpt_txt, args.char_nums)

        if (footer_txt := self.trimmed_chars_txt(char_range[1])):
            excerpt_txt.append('...\n\n').append(footer_txt)

        return self._config.text_highlighter(excerpt_txt)

    def extract_author(self) -> Name:
        """Extended in `Email` subclass to pull from headers etc."""
        return None

    def extract_timestamp(self) -> datetime | None:
        """Should be implemented in subclasses."""
        return None

    def file_display(self, align: JustifyMethod | None = None) -> FileDisplay:
        """Allows for proper right vs. left justify."""
        return FileDisplay(
            background_color=self._config.background_color,
            body_panel=BasePanel(
                border_style=self.border_style,
                text=self.prettified_txt,
                title=Text(f"({self.panel_title_timestamp})", style='dim') if self.panel_title_timestamp else None,
            ),
            document=self,
            file_info=self.file_id_panel,
            indent=site_config.indents.info,
            justify=align,
            margin_bottom=self.html_margin_bottom,
            subheaders=self.subheaders,
        )

    def lines_matching(self, _pattern: re.Pattern | str) -> list[MatchedLine]:
        """Find lines in this file matching a regex pattern."""
        pattern = patternize(_pattern)
        return [MatchedLine(line, i) for i, line in enumerate(self.lines) if pattern.search(line)]

    def print_truncated_to(self, truncate_to: int) -> None:
        """Temporarily set args.truncate and print."""
        tmp_args = {'whole_file': True} if truncate_to == NO_TRUNCATE else {'truncate': truncate_to}

        with temporary_args(tmp_args) as args:
            console.print(self)

    def print_untruncated(self) -> None:
        """Temporarily set args.whole_file to True and print."""
        self.print_truncated_to(NO_TRUNCATE)

    def raw_text(self) -> str:
        """Reload the raw data from the underlying file and return it."""
        if self.file_info.is_eml_file:
            with open(self.file_path, 'rb') as fp:
                return BytesParser(policy=policy.default).parse(fp).as_string()
        else:
            return self.file_path.read_text()

    def reload(self) -> Self:
        """Rebuild a new version of this object by loading the source file from disk again."""
        return type(self)(self.file_path)

    def repair_ocr_text(self, repairs: dict[str | re.Pattern, str], text: str) -> str:
        """Apply a dict of repairs (key is pattern or string, value is replacement string) to text."""
        for k, v in repairs.items():
            if isinstance(k, re.Pattern):
                text = k.sub(v, text)
            else:
                text = text.replace(k, v)

        return text

    def rich_header(self) -> Group:
        """Panel + subheaders with filename linking to raw file plus any additional info about the file."""
        padded_info = [Padding(sentence, site_config.info_padding()) for sentence in self.subheaders]
        return Group(*([self.file_id_panel] + padded_info))

    def top_lines(self, n: int = 10) -> str:
        """First n lines."""
        return '\n'.join(self.lines[0:n])[:DEFAULT_TRUNCATE_TO]

    def to_html(self) -> str:
        # TODO: this does not include the timestamp for OtherFiles!
        return self.file_display().to_html()

    def trimmed_chars_txt(self, truncate_to: int) -> Text | None:
        """String with link to source URL that will replace the text after the truncation point."""
        if truncate_to >= len(self.text):
            return None
        else:
            msg = f"trimmed to {truncate_to:,} characters of {self.length:,}, read the rest at {self.external_link_markup}"
            return snip_msg_txt(msg)

    def truthy_props(self, prop_names: list[str]) -> DebugDict:
        """Return key/value pairs but only if the value is truthy."""
        return {prop: getattr(self, prop) for prop in prop_names if getattr(self, prop)}

    def _coerce_entities(self, _arg: EntityScanArg) -> list[Entity]:
        """Turn strings into Entity objects."""
        return get_entities(Entity.coerce_entity_names(_arg), self)

    def _debug_dict(self, as_txt: bool = False, with_prefixes: bool = True) -> DebugDict | Text:
        """Merge information about this document from `DocCfg`, `FileInfo`, etc. objs."""
        config_info = self.config.truthy_props if self.config else {}
        file_info = self.file_info.as_dict

        # Remove duplicate fields to save space
        if config_info.get('id') == file_info.get('file_id'):
            del config_info['id']

        if file_info.get('source_url') == file_info.get('external_url', 'blah'):
            del file_info['external_url']

        if with_prefixes:
            config_info = prefix_keys(type(self.config).__name__, config_info)
            file_info = prefix_keys(underscore(FileInfo.__name__), file_info)

        props = {**config_info, **file_info, **self._debug_props()}
        return styled_dict(props) if as_txt else props

    def _debug_props(self) -> DebugDict:
        """Collects props of this object only (not the config or locations)."""
        props = {k: getattr(self, k) for k in DEBUG_PROPS}
        props.update(self.truthy_props(DEBUG_PROPS_TRUTHY_ONLY))
        props['entities'] = [e.name for e in self.entities]

        if self.file_info.file_size > 100 * 1024:
            props['file_size_str'] = self.file_info.file_size_str

        return prefix_keys(self._debug_prefix, props)

    def _debug_txt(self) -> Text:
        """Prettified version of `self._debug_dict()` suitable for printing."""
        txt_lines = styled_dict(self._debug_dict(), sep=': ')
        return prefix_with(txt_lines, ' ', pfx_style='grey', indent=2)

    def _inject_line_numbers(self, text: TextVar, interval: int) -> TextVar:
        """Inject character numbers markers into `text`. For debugging only."""
        idx = interval
        new_text = text[:idx]

        def line_marker(i: int) -> TextVar:
            marker_str = f'\n\n ------ {idx} ------ \n\n'
            return Text(marker_str) if isinstance(text, Text) else marker_str

        while idx < len(text):
            new_text += line_marker(idx)
            end_idx = idx + interval
            new_text += text[idx:end_idx]
            idx = end_idx

        return new_text

    def _load_file(self) -> str:
        """Remove BOM and HOUSE OVERSIGHT lines, strip whitespace."""
        return self.raw_text()

    def _log_top_lines(self, n: int = 10, msg: str = '', level: int = logging.DEBUG) -> None:
        """Log first 'n' lines of self.text at 'level'. 'msg' can be optionally provided."""
        separator = '\n\n' if '\n' in msg else '. '
        msg = (msg + separator) if msg else ''
        self._log(f"{msg}First {n} lines of {self.file_id}:\n\n{self.top_lines(n)}\n", level)

    def _numbered_lines(self) -> str:
        """For logging."""
        return '\n'.join([f"[{i}] {quote(line)}" for i, line in enumerate(self.lines)])

    def _repair(self) -> None:
        """Can optionally be overloaded in subclasses to further improve self.text."""
        text = self.repair_ocr_text(OCR_REPAIRS, self.text.lstrip('\ufeff').strip())  # remove BOM

        lines = [
            line.strip() if self.STRIP_WHITESPACE else line for line in text.split('\n')
            if not (line.startswith(HOUSE_OVERSIGHT) or line.startswith(EFTA_PREFIX))
        ]

        self._set_text(text=collapse_newlines('\n'.join(lines)))

    def _set_text(self, lines: list[str] | None = None, text: str | None = None) -> None:
        """Set `self.text` and `self.lines` based on arguments passed."""
        if lines and text:
            raise RuntimeError(f"[{self.filename}] Either 'lines' or 'text' arg must be provided (got both)")
        elif lines is not None:
            self.text = '\n'.join(lines).strip()
        elif text is not None:
            self.text = text.strip()
        else:
            raise RuntimeError(f"[{self.filename}] Either 'lines' or 'text' arg must be provided (neither was)")

        # logger.debug(f"_set_text() set self.text to\n---\n{self.text}\n---")
        self.lines = [line.strip() if self.STRIP_WHITESPACE else line for line in self.text.split('\n')]

    def _skipped_file_txt(self, reason: str | Text) -> Text:
        txt = Text(f"Skipping ", f"{INFO_STYLE} dim").append(self.file_info.external_link_txt(self.author_style))
        return txt.append(" because it's ").append(reason)

    def _intro_txt(self, cutoff: int) -> Text:
        """Truncation message if it's an excerpt."""
        if cutoff == 0:
            return Text('')  # Empty Text object makes sure the whole string starts with default no-style

        return snip_msg_txt(f'trimmed first {cutoff:,} characters', EXCERPT_STYLE).append('\n\n...')

    def _write_clean_text(self, output_path: Path) -> None:
        """Write self.text to 'output_path'. Used only for diffing files."""
        if output_path.exists():
            if str(output_path.name).startswith(HOUSE_OVERSIGHT_PREFIX):
                raise RuntimeError(f"'{output_path}' already exists! Not overwriting.")
            else:
                logger.warning(f"Overwriting '{output_path}'...")

        with open(output_path, 'w') as f:
            f.write(self.text)
            self._log(f"Wrote {self.length} chars of self.text to '{output_path}'.")

    @contextmanager
    def _write_tmp_file(self) -> Generator[Path, None, None]:
        with NamedTemporaryFile(dir=self.file_path.parent) as tmp_doc_file:
            tmp_path = Path(tmp_doc_file.name)
            self._write_clean_text(tmp_path)
            self._log(f"created tmp file '{tmp_doc_file.name}' ({file_size_str(tmp_path)})")
            yield Path(tmp_doc_file.name)

    def __rich_console__(self, console: Console, options: ConsoleOptions) -> RenderResult:
        """Default `Document` renderer (Email and MessengerLog override this)."""
        yield self.file_display()

    def __str__(self) -> str:
        return self._summary.plain

    @classmethod
    def default_category(cls) -> str:
        return ''

    @classmethod
    def dummy_cfg(cls) -> DocCfg:
        """Empty config so you can call config methods without "if self.config and self.config.thing"."""
        return DocCfg(id='DUMMY')

    @classmethod
    def files_summary_table(cls, title: str | Text, first_col_name: str) -> Table:
        """Empty table with appropriate cols for summarizing groups of files."""
        table = build_table(title)
        cols = [{'name': first_col_name, 'min_width': 14}] + SUMMARY_TABLE_COLS
        add_cols_to_table(table, cols, 'right')
        return table

    @classmethod
    def files_summary(cls, files: Sequence[Self], is_author_na: bool = False) -> dict[str, str | Text]:
        """Summary info about a group of files."""
        file_count = len(files)
        author_count = cls.known_author_count(files)

        # NOTE: Order matters!
        return {
            'count': str(file_count),
            'author_count': NA_TXT if is_author_na else str(author_count),
            'no_author_count': NA_TXT if is_author_na else str(file_count - author_count),
            'uncertain_author_count': NA_TXT if is_author_na else str(len([f for f in files if f._config.author_uncertain])),
            'bytes': file_size_to_str(sum([f.file_info.file_size for f in files])),
        }

    @classmethod
    def file_summary_row(cls, files: Sequence[Self], author_na: bool = False) -> Sequence[str | Text]:
        """Turn the values in the `cls.files_info()` dict into a list so they can be used as a table row."""
        return [v for v in cls.files_summary(files, author_na).values()]

    @classmethod
    def filter_for_type(cls, docs: Sequence['Document']) -> list[Self]:
        """Filter for Document objects of this class."""
        return [d for d in docs if isinstance(d, cls)]

    @classmethod
    def known_author_count(cls, docs: Sequence[Self]) -> int:
        """Number of elements of `docs` that have an author attribution."""
        return len([doc for doc in docs if doc.author])

    @classmethod
    def _print_ids(cls, docs: Sequence[Self], msg: str = '') -> None:
        """Debug method to print raw string of IDs suitable for copy/paste."""
        print(f"\n\n IDs for {msg}:\n\n{' '.join([doc.file_id for doc in docs])}\n")

    ########################
    #### STATIC METHODS ####
    ########################

    @staticmethod
    def count_by_month(docs: Sequence['DocType']) -> Counter[str | None]:
        return Counter([d.timestamp.date().isoformat()[0:7] if d.timestamp else None for d in docs])

    @staticmethod
    def sort_by_id(docs: Sequence['DocType']) -> list['DocType']:
        return sorted(docs, key=lambda d: d.file_id)

    @staticmethod
    def sort_by_length(docs: Sequence['DocType']) -> list['DocType']:
        return sorted(docs, key=lambda d: d.file_info.file_size, reverse=True)

    @staticmethod
    def sort_by_timestamp(docs: Sequence['DocType']) -> list['DocType']:
        return sorted(docs, key=lambda doc: doc.timestamp_sort_key)

    @staticmethod
    def uniquify(documents: Sequence['DocType']) -> Sequence['DocType']:
        """Uniquify by file_id."""
        id_map = {doc.file_id: doc for doc in documents}
        return [doc for doc in id_map.values()]

    @staticmethod
    def without_dupes(docs: Sequence['DocType']) -> list['DocType']:
        """Remove any duplicate documents."""
        return [doc for doc in docs if not doc.is_duplicate]


DocType = TypeVar('DocType', bound=Document)



import json
import re
from dataclasses import asdict, dataclass, field

from rich.text import Text

from epstein_files.output.epstein_highlighter import highlighter


@dataclass
class EmailParts:
    """Simple container to hold the header and body strings for an email in separate variables."""
    header: str
    body: str

    def __post_init__(self):
        self.header = self.header.strip()
        self.body = self.body.strip()

    @property
    def header_txt(self) -> Text:
        return highlighter(self.header)

    @property
    def header_len(self) -> int:
        return len(self.header)

    def __str__(self) -> str:
        return f"{self.header}\n\n{self.body}"



"""
Constants used to build internal links within one html page.
"""
from enum import auto, StrEnum


AUTHORS_USING_SIGNATURES = 'Authors Seen Using Email Signatures'
SELECTIONS_FROM = 'Selections from '
HIS_EMAILS = 'His Emails'
HIS_TEXT_MESSAGES = 'His Text Messages'
FILES_THAT_ARE_NEITHER_EMAILS_NOR = 'Files That Are Neither Emails Nor Text Messages'
TO_FROM = 'to/from'


class PageSections(StrEnum):
    EMAILS = auto()
    EMAIL_SIGNATURES = auto()
    OTHER_FILES = auto()
    TEXT_MESSAGES = auto()


# Search terms that take you to the desired section via magic URL comment arg
SECTION_ANCHORS = {
    PageSections.EMAILS: SELECTIONS_FROM + HIS_EMAILS,
    PageSections.EMAIL_SIGNATURES: AUTHORS_USING_SIGNATURES,
    PageSections.TEXT_MESSAGES: SELECTIONS_FROM + HIS_TEXT_MESSAGES,
    PageSections.OTHER_FILES: FILES_THAT_ARE_NEITHER_EMAILS_NOR,
}



from dataclasses import asdict, dataclass, field, fields

from epstein_files.documents.config.doc_cfg import SHORT_TRUNCATE_TO, CharRange
from epstein_files.documents.config.communication_cfg import CommunicationCfg, Platform


@dataclass(kw_only=True)
class EmailCfg(CommunicationCfg):
    """
    Attributes:
        actual_text (str, optional): In dire cases of broken OCR we just configure the body of the email as a string.
        fwded_text_after (str, optional): If set, any text after this is a fwd of an article or similar.
        has_uninteresting_ccs (bool): If `True` this email's CC: recipients will be marked as 'uninteresting'.
        has_uninteresting_bccs (bool): If `True` this email's BCC: recipients will be marked as 'uninteresting'.
        subject (str, optional): Email subject line.
    """
    platform: Platform = Platform.EMAIL
    actual_text: str | None = None
    fwded_text_after: str | None = None
    has_uninteresting_ccs: bool = False
    has_uninteresting_bccs: bool = False
    subject: str | None = None

    def __post_init__(self):
        super().__post_init__()

        if self.fwded_text_after:
            self.is_valid_for_name_scan = False

    @property
    def truncate_at(self) -> int | CharRange | None:
        if super().truncate_at:
            return super().truncate_at
        elif self.is_fwded_article or self.fwded_text_after:
            return SHORT_TRUNCATE_TO

    @property
    def truthy_props(self) -> dict[str, bool | str | None]:
        props = super().truthy_props

        if props.get('platform') == Platform.EMAIL:
            del props['platform']

        return props

    # This is necessary because for some dumb reason @dataclass(repr=False) doesn't cut it
    def __repr__(self) -> str:
        return super().__repr__()



import re
import logging
import warnings
from dataclasses import asdict, dataclass
from datetime import datetime, timezone
from typing import ClassVar, Sequence

import datefinder
import dateutil
from rich.console import Group
from rich.markup import escape
from rich.panel import Panel
from rich.table import Table
from rich.text import Text

from epstein_files.documents.document import CLOSE_PROPERTIES_CHAR, Document
from epstein_files.documents.config.config_builder import build_cfg_from_text
from epstein_files.documents.config.doc_cfg import DocCfg, Metadata
from epstein_files.documents.documents.file_info import FileInfo
from epstein_files.output.epstein_highlighter import highlighter
from epstein_files.output.highlight_config import styled_category
from epstein_files.output.rich import build_table, console
from epstein_files.people.interesting_people import PERSONS_OF_INTEREST
from epstein_files.util.constant.strings import *
from epstein_files.util.constants import *
from epstein_files.util.helpers.data_helpers import days_between, coerce_utc_strict, uniquify, uniq_sorted
from epstein_files.util.helpers.debugging_helper import tz_debug_str
from epstein_files.util.helpers.file_helper import FILENAME_LENGTH
from epstein_files.util.helpers.rich_helpers import QUESTION_MARKS_TXT
from epstein_files.util.helpers.string_helper import DATE_LENGTH, collapse_whitespace, indented
from epstein_files.util.env import args, site_config
from epstein_files.util.logging import logger

FIRST_FEW_LINES = 'First Few Lines'
MAX_DAYS_SPANNED_TO_LOG_TOP_LINES = 10
MAX_EXTRACTED_TIMESTAMPS = 100
MIN_TIMESTAMP = coerce_utc_strict(datetime(2000, 1, 1))
MIN_PAGES_TO_TRUNCATE_PREVIEW = 10
TRUNCATED_PREVIEW_LEN = 200

# There's 3 spaces (2 for padding, one for divider) between each col and then 2 on each side + 1 indent
PREVIEW_COL_WIDTH = console.width - FILENAME_LENGTH - DATE_LENGTH - (3 * 2) - 6

SKIP_TIMESTAMP_EXTRACT = [
    PALM_BEACH_TSV,
    PALM_BEACH_PROPERTY_INFO,
]

# TODO: get rid of this
VAST_HOUSE = 'vast house'  # Michael Wolff article draft about Epstein indicator


@dataclass
class OtherFile(Document):
    """
    File that is not an email, an iMessage log, or JSON data.

    Attributes:
        derived_cfg (DocCfg, optional): `DocCfg` object derived from contents of the file instead of being configured
    """
    derived_cfg: DocCfg | None = None

    # Class vars
    _INCLUDE_DESCRIPTION_IN_SUMMARY_PANEL: ClassVar[bool] = True                   # Overrides superclass
    MAX_TIMESTAMP: ClassVar[datetime] = coerce_utc_strict(datetime(2022, 12, 31))  # Overloaded in DojFile

    def __post_init__(self):
        super().__post_init__()

        if not self.config and (cfg := build_cfg_from_text(self)):
            self.derived_cfg = cfg

    @property
    def config(self) -> DocCfg | None:
        return super().config or self.derived_cfg

    @property
    def is_interesting(self) -> bool | None:
        """Junk emails are not interesting."""
        if (is_interesting := super().is_interesting) is not None:
            return is_interesting
        elif self.author is None and self.file_info.is_house_oversight_file:
            return True
        elif self.author is not None and self.author in PERSONS_OF_INTEREST:
            return True

    @property
    def is_valid_for_table(self) -> bool:
        """Return True if this file is OK to put in a table in the curated chronological views."""
        return not (self._config.is_excerpt or self._config.show_full_panel)

    @property
    def metadata(self) -> Metadata:
        metadata = super().metadata
        metadata['is_interesting'] = self.is_interesting

        if self.extracted_timestamp:
            metadata['is_timestamp_inferred'] = True

        return metadata

    @property
    def preview_chars(self) -> str:
        """Text at start of file stripped of newlines for display in tables and other cramped settings."""
        if self._config.replacement_preview_text:
            return self._config.replacement_preview_text

        # TODO: config.num_preview_chars is never used
        num_chars = self._config.num_preview_chars or site_config.other_files_preview_chars
        text = self.display_text[0:num_chars]

        if text.count('Page') > MIN_PAGES_TO_TRUNCATE_PREVIEW:
            num_chars = TRUNCATED_PREVIEW_LEN

        return collapse_whitespace(text)[0:num_chars]

    @property
    def preview_txt(self) -> Text:
        if self._config.display_preview_txt and not self._config.has_full_ocr_text_replacement:
            return self._config.display_preview_txt

        txt = highlighter(escape(self.preview_chars))

        # TODO: should check self.length > len(self.preview_chars) but won't quite work with prettified text insertions
        if self.length > site_config.other_files_preview_chars:
            txt.append(f"... ({self.length - len(txt):,} more characters)", 'dim italic')

        return txt

    @property
    def _summary(self) -> Text:
        """One line summary mostly for logging."""
        return super()._summary.append(CLOSE_PROPERTIES_CHAR)

    def extract_timestamp(self) -> datetime | None:
        """Return configured timestamp or value extracted by scanning text with datefinder."""
        timestamps: list[datetime] = []

        # NOTE: this is a lame optimization for speed
        if any([s in self._config.complete_description for s in SKIP_TIMESTAMP_EXTRACT]):
            return None

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", module="dateutil")

            try:
                # TODO: datefinder.find_dates() cannot find 08/29/2019 style e.g. in EFTA00005783 :(
                for dt in datefinder.find_dates(self.text, strict=False):
                    if MIN_TIMESTAMP < (dt := coerce_utc_strict(dt)) < self.MAX_TIMESTAMP:
                        timestamps.append(dt)

                    # Stop find_dates() once we've located a decent number of timestamps
                    if len(timestamps) >= MAX_EXTRACTED_TIMESTAMPS:
                        break
            except ValueError as e:
                self._warn(f"{e}\nError iterating over dateutil.find_dates(), using {len(timestamps)} timestamps found so far")

        return self._choose_extracted_timestamp(timestamps)

    def _choose_extracted_timestamp(self, timestamps: list[datetime]) -> datetime | None:
        """Most recent timestamp appearing in the text is usually the most likely to be correct."""
        timestamps = uniq_sorted(timestamps, reverse=True)

        if len(timestamps) == 0:
            if not (self.is_duplicate or VAST_HOUSE in self.text):
                self._log_top_lines(15, msg=f"No timestamps found")

            return None

        timestamp = timestamps[0]
        days_spanned = days_between(timestamps[-1], timestamp)

        if len(timestamps) == 1:
            log_msg = f"Found only one extracted timestamp '{timestamp}', returning it..."
        else:
            log_msg = f"Choosing '{timestamp}' from {len(timestamps)} extracted timestamps " \
                      f"spanning {days_spanned} days\n\n" + \
                      indented([tz_debug_str(dt) for dt in timestamps])

        if days_spanned > MAX_DAYS_SPANNED_TO_LOG_TOP_LINES:
            self._log_top_lines(10, msg=log_msg, level=logging.DEBUG)
        else:
            self._debug_log(log_msg)

        return timestamp

    @classmethod
    def files_preview_table(
        cls,
        files: Sequence['OtherFile'],
        title_pfx: str = '',
        title: str | Text | None = '',
        title_justify: str = '',
        **kwargs
    ) -> Table:
        """Build a table of `OtherFile` documents."""
        title = '' if title is None else (title or f'{title_pfx}Other Files Details in Chronological Order')
        title_justify = title_justify or ('left' if title else 'center')
        table = build_table(title, show_lines=True, title_justify=title_justify, **kwargs)
        table.add_column('File', justify='center', width=FILENAME_LENGTH)
        table.add_column('Info', justify='center', width=DATE_LENGTH)
        table.add_column(FIRST_FEW_LINES, justify='left', style='pale_turquoise4', width=PREVIEW_COL_WIDTH)

        for file in files:
            # call superclass method to avoid border_style rainbow
            link_and_info = [FileInfo.build_external_links(file.file_info, file.category_style, id_only=bool(args.mobile))]

            if file.date_str:
                date_txt = Text(file.date_str, style=TIMESTAMP_STYLE)
            else:
                date_txt = Text('(no date)', style=f"{TIMESTAMP_STYLE} dim italic")

            if file.is_duplicate:
                preview_text = file.duplicate_file_txt
                row_style = 'dim'
            else:
                preview_text = file.preview_txt
                row_style = ''
                link_and_info += file.subheaders

            table.add_row(
                Group(*link_and_info),
                Group(date_txt, file._config.category_txt, Text(file.file_info.file_size_str, style='wheat4 dim')),
                preview_text,
                style=row_style
            )

        return cls._mobilize_table(table) if args.mobile else table

    @classmethod
    def _mobilize_table(cls, _table: Table) -> Table:
        """Make a mobile version of the `files_preview_table()`."""
        table = build_table(_table.title, copy_props_from=_table)
        table.add_column('File', justify='center', width=len('EFTA00424931'))
        table.add_column(FIRST_FEW_LINES, justify='left', style='pale_turquoise4')
        max_col_idx = len(_table.columns) - 1

        for i, row in enumerate(_table.rows):
            table.add_row(
                # Collapse all but last col into one
                Group(
                    *[_table.columns[j]._cells[i] for j in range(1, max_col_idx)],
                    _table.columns[0]._cells[i],
                ),
                _table.columns[max_col_idx]._cells[i],
                style=row.style
            )

        return table

    @classmethod
    def summary_table(cls, files: Sequence['OtherFile'], title_pfx: str = '') -> Table:
        """Table showing file count by category."""
        categories = uniquify([f.category for f in files])
        categories = sorted(categories, key=lambda c: -len([f for f in files if f.category == c]))
        table = cls.files_summary_table(f'{title_pfx}Other Files Summary', 'Category')

        for category in categories:
            category_files = [f for f in files if f.category == category]
            table.add_row(styled_category(category), *cls.file_summary_row(category_files))

        table.columns = table.columns[:-2] + [table.columns[-1]]  # Removee unknown author col
        return table



import logging
import re
from dataclasses import dataclass, field, fields
from typing import ClassVar, Literal, Self, Sequence

from rich.padding import Padding
from rich.style import Style
from rich.text import Text

from epstein_files.output.html.rich_style import RichStyle
from epstein_files.people.names import UNKNOWN, Name, constantize_name, extract_first_name, extract_last_name
from epstein_files.util.constant.strings import INDENT_NEWLINE, INDENTED_JOIN, LAW_ENFORCEMENT, WIKIPEDIA, PartialName
from epstein_files.util.constant.urls import EPSTEINIFY, PERSON_LINK_BUILDERS, EpsteinSite, wikipedia_url
from epstein_files.util.env import args, site_config
from epstein_files.util.external_link import ExternalLink, link_text_obj
from epstein_files.util.helpers.data_helpers import constantize_names, listify, without_falsey
from epstein_files.util.helpers.rich_helpers import QUESTION_MARKS_TXT, Textish, enclose, join_texts, parenthesize
from epstein_files.util.helpers.string_helper import (as_pattern, indented, is_integer, join_patterns,
     join_truthy, quote, remove_question_marks)
from epstein_files.util.logging import logger
from epstein_files.util.logging_entity import LoggingEntity
from epstein_files.util.constant.urls import internal_person_link_url

AKA_STYLE = 'grey39 italic'
BIO_COLOR = 'grey70'
BIO_STYLE = f'italic {BIO_COLOR}'
MIN_LEN_FOR_OPTIONAL_LAST_CHAR = 5

LINK_JOIN_STYLE = 'grey23 bold'
MGMT_PATTERN = r"M(ana)?ge?m(en)?t"
MGMT_REGEX = re.compile(MGMT_PATTERN)
COMPANY_SUFFIX_REGEX = re.compile(fr".*?(,? (Inc\.?|LLC|{MGMT_PATTERN}))$")
MIDDLE_INITIAL_REGEX = re.compile(r"^[A-Z]\.?$")
SIMPLE_NAME_REGEX = re.compile(r"^[-\w, ]+$", re.IGNORECASE)


@dataclass
class Entity(LoggingEntity):
    """
    Attributes:
        name (str): Person or organization name
        info (str, optional): biographical info about this person
        emailer_pattern (str, optional): manually constructed regex pattern to match this person in email headers
        aliases (list[str]): known aliases
        category (str, optional): category of this entity
        email_addresses (list[str]): known email addresses
        emailer_regex (re.Pattern): pattern that matches this person's name in email headers
        highlight_regex (re.Pattern): pattern that matches this person's name in various variations for highlightihng
        is_emailer (bool): should email headers be scanned for this entity
        is_interesting (bool): should a biographical entry be generated for this panel in the chronological view
        is_junk (bool): for junk email
        match_partial (PartialName | None): whether to also match this entity's first and/or last names
        style (str, optional): style to use when printing this entity's name
        unique_phraseologies (list[str], optional): turns of phrase known to be unique to this person
        url (str | list[str] | Literal['WIKIPEDIA'], optional): link(s) to info about this entity
    """
    name: str
    info: str = ''  # TODO: rename "note" for consistency w/DocCfg
    emailer_pattern: str = ''

    # Props after here not set by positional args
    aliases: list[str] = field(default_factory=list)
    email_addresses: list[str] = field(default_factory=list)
    emailer_regex: re.Pattern = field(init=False)
    highlight_regex: re.Pattern = field(init=False)
    is_emailer: bool | None = True
    is_interesting: bool = True  # Eligible for bio panel
    is_junk: bool = False  # TODO: this sucks
    match_partial: PartialName | None = 'last'
    unique_phraseologies: list[str] = field(default_factory=list)
    url: str | list[str] | Literal['WIKIPEDIA'] = ''

    # NOTE: category is usually set by a `HighlightedNames` that holds this `Entity`
    category: str = ''
    _style: RichStyle = field(default_factory=lambda: RichStyle(None))
    _urls: list[str] = field(init=False)
    # jmail_url: str

    # Class variables
    DEFAULT_PATTERN_SFX: ClassVar[str] = '?'

    @property
    def style(self) -> str:
        style_str = str(self._style.style)
        return '' if style_str == 'none' else style_str

    @style.setter
    def style(self, val: str | Style | None):
        self._style = RichStyle(val)

    def __post_init__(self):
        self._urls = [wikipedia_url(self.name) if url == WIKIPEDIA else url for url in listify(self.url)]

        try:
            self.emailer_regex = re.compile(self.pattern, re.IGNORECASE)
            self.highlight_regex = re.compile(fr"\b({self.highlight_pattern})\b", re.IGNORECASE)
        except re.error as e:
            self._log(f"failed to compile emailer or highlight regex: {e}", logging.ERROR)
            raise e

        if self.category:
            self._warn(f"has category '{self.category}' at instantiation time (style='{self.style}')")

    @property
    def alt_links(self) -> list[ExternalLink]:
        """Links beyond the first one (which is usually attached to the name)."""
        return [ExternalLink(url, f'more', link_style=self.style) for i, url in enumerate(self._urls[1:], 2)]

    @property
    def alt_links_txt(self) -> Text:
        """Alternate links parenthesized and concatenated into one Text object."""
        if self.alt_links:
            return enclose(join_texts(self.alt_links, Text('/', LINK_JOIN_STYLE)), '()', LINK_JOIN_STYLE)
        else:
            return Text('')

    # TODO: add known email addresses?
    @property
    def bio_txt(self) -> Text:
        """Biographical info about this entity with links etc."""
        from epstein_files.output.epstein_highlighter import non_epstein_highlighter
        bio_pieces: list[Textish] = [Text('').append(self.name_with_link)]
        bio_pieces.extend([Text('').append('aka ', AKA_STYLE).append(Text(alias, self.style)) for alias in self.aliases])

        if self.category:
            category_txt = Text(self.category.lower(), style=self._style.dim)
            bio_pieces.append(enclose(category_txt, encloser='[]', encloser_style='dim'))

        bio_pieces.append(non_epstein_highlighter(Text(self.info, BIO_STYLE)) if self.info else QUESTION_MARKS_TXT)
        bio_pieces.append(self.alt_links_txt)
        return join_texts(bio_pieces)

    # TODO: rename this somehting that means "non_custom_external_links"
    @property
    def epstein_sites_all_links(self) -> Text:
        """Collection of links to sites that have biographies of the Epstein network like epsteinify, Jmail search, etc."""
        links = [self.epstein_site_link(site) for site in PERSON_LINK_BUILDERS]
        return Text('', justify='center', style='dim').append(join_texts(links, join=' / '))

    @property
    def has_bio(self) -> bool:
        return bool(self.is_interesting and self.info)

    @property
    def highlight_pattern(self) -> str:
        """`self.emailer_pattern` extended with first/last name variations for Rich highlighting."""
        if self.is_junk:
            return self.pattern  # TODO: this sucks

        return join_patterns(self._name_patterns)

    @property
    def identifying_strings(self) -> list[str]:
        """Strings that indicate a document is likely tied to this entity."""
        return self.unique_phraseologies + (self.email_addresses if self.is_emailer else [])

    @property
    def info_with_category(self) -> str:
        """String lik 'category, info'."""
        return ', '.join(without_falsey([self.category, self.info]))

    @property
    def internal_link(self) -> Text:
        """Kind of like an anchor link to the section of the page containing these emails."""
        return link_text_obj(internal_person_link_url(self.name), self.name, self.style)

    @property
    def is_email_address(self) -> bool:
        return '@' in self.name

    @property
    def is_linkable(self) -> bool:
        """Return True if it's likely that EpsteinWeb has a page for this name."""
        return not (self.is_email_address or '/' in self.name)

    @property
    def name_with_link(self) -> Text:
        """Colored name with hyperlink if applicable (otherwise just colored name)."""
        if self._urls:
            return ExternalLink(self._urls[0], self.name, link_style=self._style.bold).link
        else:
            return Text(self.name, self._style.not_bold)

    @property
    def pattern(self) -> str:
        """Pattern used for matching emails, base pattern for highlights."""
        if self.emailer_pattern:
            pattern = self.emailer_pattern
        else:
            if self._middle_initial:
                pattern = fr"{self._names[0]} ({self._middle_initial}\.? )?{self._names[-1]}"
            else:
                pattern = self.name

            if len(self.name) >= MIN_LEN_FOR_OPTIONAL_LAST_CHAR:
                pattern += self.DEFAULT_PATTERN_SFX

        return as_pattern(join_patterns([pattern, *self.email_addresses]))

    @property
    def _identifier(self) -> str:
        """`LoggingEntity` abstract method implementation."""
        return quote(self.name)

    @property
    def _middle_initial(self) -> str:
        """Returns 'J' for 'Albert J. Gomez'."""
        if len(self._names) != 3:
            return ''
        elif MIDDLE_INITIAL_REGEX.match(self._names[1]):
            return self._names[1][0]
        else:
            return ''

    @property
    def _name_patterns(self) -> list[str]:
        """['Firstname', 'Lastname', 'Lastname, Firstname'."""
        name_patterns = [self.pattern]
        name = remove_question_marks(self.name)  # TODO: this sucks

        if self.match_partial is None or ' ' not in self.name:
            return name_patterns

        first_name = extract_first_name(name)
        last_name = extract_last_name(name)
        name_patterns.append(as_pattern(f"{last_name},? {first_name}"))  # TODO: maybe reversed name should be included by default?

        if self.match_partial in ['both', 'first'] and SIMPLE_NAME_REGEX.match(first_name):
            name_patterns.append(as_pattern(first_name))

        if self.match_partial in ['both', 'last'] and SIMPLE_NAME_REGEX.match(last_name):
            name_patterns.append(as_pattern(last_name))

        if args._debug_highlight_patterns:
            self._debug_log(f"name_patterns: '{name_patterns}'")

        return name_patterns

    @property
    def _names(self) -> list[str]:
        return self.name.split()

    @property
    def _props_strs(self) -> list[str]:
        props = []
        # if there's no self.emailer_pattern this obj was probably instantiated with positional args only
        add_prop = lambda f, value: props.append(f"{f}={value}" if self.emailer_pattern else value)

        for _field in fields(self):
            if _field.name == 'name':
                add_prop(_field.name, constantize_name(getattr(self, _field.name)))
            elif _field.name == '_style':
                add_prop('style', quote(self.style))
            elif (value := getattr(self, _field.name)) and not _field.name.endswith('regex'):
                if _field.name.endswith('pattern'):
                    value = f'r"{value}"'
                else:
                    value = quote(constantize_names(str(value)))
                    value = ('f' + value) if '{' in value else value

                add_prop(_field.name, value)

        props.append(f'highlight_pattern=r"{self.highlight_pattern}"')
        return props

    @property
    def _style_bold(self) -> Style:
        return self._style.not_bold

    @property
    def _style_no_bold(self) -> Style:
        return self._style.bold

    @property
    def _style_dim(self) -> str:
        return self.style if 'dim' in self.style else f'{self.style} dim'

    def epstein_site_link(self, site: EpsteinSite = EPSTEINIFY, link_text: str | None = None) -> ExternalLink:
        """Link to epsteinify.com, EpsteinWeb, or similar for this name."""
        return ExternalLink(self.epstein_site_url(site), link_text or site, link_style=self.style)

    def epstein_site_url(self, site: EpsteinSite = EPSTEINIFY) -> str:
        """URL pointing to info about this entity on one of the Epstein sites like epsteinify.com."""
        return PERSON_LINK_BUILDERS[site](self.name)

    def __eq__(self, other: Self):
        if not isinstance(other, Self):
            return NotImplemented

        return self.name == other.name

    def __hash__(self):
        return hash(self.name)

    def __repr__(self) -> str:
        props = self._props_strs
        type_str = f"{type(self).__name__}("

        if self.emailer_pattern:
            repr_str = f"{type_str}{INDENT_NEWLINE}" + INDENTED_JOIN.join(props)
            repr_str += ',' if props else ''
            repr_str += '\n)'
        else:
            repr_str = type_str + ', '.join(props) + ')'

        return repr_str

    def __str__(self) -> str:
        return self.name

    @classmethod
    def coerce_entity_names(cls, _arg: 'EntityScanArg') -> list[str]:
        if not _arg:
            return []

        return [e.name if isinstance(e, cls) else e for e in listify(_arg)]

    @classmethod
    def _repr_string(cls, contact_infos: list[Self]) -> str:
        """Print a list of `Entity` objects to a python string that can recreate them when executed."""
        return '[\n' + indented([repr(contact) for contact in contact_infos], 4) + '\n],'


@dataclass(eq=False)
class Organization(Entity):
    # Different defaults
    is_emailer: bool | None = None
    match_partial: PartialName | None = None
    # Additional properties
    belongs_to: str = ''

    DEFAULT_PATTERN_SFX: ClassVar[str] = ''

    def __post_init__(self):
        if self.emailer_pattern and self.is_emailer is not False:
            self.is_emailer = True

        if (suffix_match := COMPANY_SUFFIX_REGEX.match(self.name)) and not self.emailer_pattern:
            suffix = suffix_match.group(1)
            emailer_pattern = self.name.removesuffix(suffix)

            if not MGMT_REGEX.search(suffix):
                if suffix.startswith(','):
                    suffix = suffix.replace(',', ',?')
                else:
                    suffix = f",?{suffix}"

            if suffix.endswith('.'):
                suffix = suffix.replace('.', r'\.?')

            self.emailer_pattern += fr"{emailer_pattern}({suffix})?"

        if self.belongs_to:
            if self.info:
                self.info = join_truthy(f"{self.belongs_to}", self.info)
            else:
                self.info = f"{self.belongs_to} organization"

        return super().__post_init__()

    @classmethod
    def well_known(cls, name: str, **kwargs) -> Self:
        """Alternate constructor that sets is_interesting to False."""
        return cls(name, is_interesting=False, **kwargs)


EntityScanArg = Sequence[Entity | str] | Entity | str | None


def acronym(name: str, info: str = '', **kwargs) -> Organization:
    """Auto-generates a regex matching the org's initials."""
    initials = [word[0] for word in name.split() if word[0].isupper()]
    initials_pattern = ''.join([fr"{letter}\.?" for letter in initials])
    emailer_pattern = join_patterns([initials_pattern, name])

    # Force word boundary in front to prevent spurious matches (can't force at end bc of period)
    if (is_emailer := _pop_kwarg(kwargs, 'is_emailer', False)):
        emailer_pattern = fr"\b({emailer_pattern})"

    return Organization(
        ''.join(initials),
        join_truthy(name, info, ', '),
        emailer_pattern,
        is_emailer=is_emailer,
        **kwargs
    )


def epstein_co(name: str, emailer_pattern: str = '', info: str = '', **kwargs) -> Organization:
    return Organization(name, join_truthy('Epstein company', info), emailer_pattern, **kwargs)


def epstein_trust(
    name: str,
    emailer_pattern: str = '',
    beneficiaries: list[str] | None = None,
    trustees: list[str] | None = None,
) -> Organization:
    """One of Epstein's financial trust entities."""
    name = f"Jeffrey E. Epstein {name} Trust" if is_integer(name) else name
    beneficiary_str = ''

    if beneficiaries:
        if len(beneficiaries) == 1:
            beneficiary_str = f"sole beneficiary {beneficiaries[0]}"
        else:
            beneficiary_str = f"beneficiaries {', '.join(beneficiaries)}"

    if trustees:
        beneficiary_str = join_truthy(beneficiary_str, f"trustees: " + ', '.join(trustees), ', ')

    beneficiary_str = f", {beneficiary_str}" if beneficiary_str else ''
    return Organization(name, f'Epstein financial trust{beneficiary_str}', emailer_pattern)


def law_enforcement(name: str, emailer_pattern: str = '', info: str = '', **kwargs) -> Organization:
    is_interesting = _pop_kwarg(kwargs, 'is_interesting', False)
    is_emailer = _pop_kwarg(kwargs, 'is_emailer', False)

    if name.startswith('US ') and not emailer_pattern:
        emailer_pattern = r"U(\.|nited)? S(\.|tates)? " + name.removeprefix('US ')

    return Organization(
        name,
        info or LAW_ENFORCEMENT,
        emailer_pattern,
        is_emailer=is_emailer,
        is_interesting=is_interesting,
        **kwargs
    )


def publication(name: str, emailer_pattern: str = '', **kwargs) -> Organization:
    """Convenience method for WSJ, New York Times, etc."""
    # Make sure not to match 'Daily News' to 'Virgin Islands Daily News' / 'Palm Beach Daily News'
    if name.startswith('Daily'):
        emailer_pattern = fr"(?<!(Beach|lands)\s){as_pattern(emailer_pattern or name)}"

    return Organization(name, emailer_pattern=emailer_pattern, is_interesting=False, **kwargs)


def the_publication(name: str, emailer_pattern: str = '', **kwargs) -> Organization:
    """Publication that starts with 'The'."""
    emailer_pattern = emailer_pattern or name
    pattern = fr"({emailer_pattern})" if '|' in emailer_pattern else emailer_pattern
    return publication(name, fr"(The )?{pattern}", **kwargs)


def _pop_kwarg(kwargs: dict, key: str, default: bool = False) -> bool:
    """Pop a kwarg and return if it exists, elese return `default`."""
    return kwargs.pop(key) if key in kwargs else default



from epstein_files.documents.config.communication_cfg import CommunicationCfg
from epstein_files.documents.config.config_builder import blaine_letter, immigration_letter, letter
from epstein_files.documents.config.email_cfg import EmailCfg
from epstein_files.people.names import *


# This category makes is_interesting default to True
LETTER_CFGS = [
    letter(
        id='026011',
        author='Gennady Mashtalyar',
        date='2016-06-24',  # date is based on Brexit reference but he could be backtesting,
        note=f"about algorithmic trading",
    ),
    letter(id='026134', recipients=['George'], note=f'about opportunities to buy banks in Ukraine'),
    blaine_letter('019086', date='2015-05-27', suffix='naming various Putin puppet regimes', show_full_panel=True),
    blaine_letter('019474', date='2015-05-29'),
    blaine_letter('019476', date='2015-06-01'),
    immigration_letter('EFTA02389172', 'Michael Harrigan'),
    immigration_letter('EFTA00314931', TERJE_ROD_LARSEN, '2017-06-08', 'speaking for the International Peace Institute', show_with_name=MASHA_DROKOVA),
    immigration_letter('EFTA00308104', MARTIN_NOWAK, '2017-06-06', f'possibly about {MASHA_DROKOVA} visa', show_with_name=MASHA_DROKOVA),
    immigration_letter('EFTA00589797', TERJE_ROD_LARSEN, '2015-05-02'),
    immigration_letter('EFTA01086449', ARIANE_DE_ROTHSCHILD),
    immigration_letter('EFTA01143800', None, show_with_name=SVETLANA_POZHIDAEVA),
    immigration_letter('EFTA00537633', MARK_ZEFF, '2012-03-19', 'about SLK Designs, someone in interior design industry'),
    EmailCfg(id='EFTA02652017', note="inquiry about someone's immigration asylum application", is_interesting=True),

    # DOJ files
    letter(id='EFTA00007609', recipients=['Alberto'], duplicate_ids=['EFTA00007582']),
    letter(id='EFTA02731023', author='Senator Ron Wyden', recipients=[LEON_BLACK], is_interesting=False),
    letter(id='EFTA02731018', author='Senator Ron Wyden', recipients=['Marc Rowan'], is_interesting=False),
]



"""
Functions that build rich Text links.
"""
import re
from dataclasses import dataclass
from typing import Self, Sequence
from urllib.parse import urlsplit

from rich.style import Style
from rich.text import Text

from epstein_files.util.constant.strings import ARCHIVE_LINK_COLOR, ARCHIVE_ALT_LINK_STYLE, ARCHIVE_LINK_COLOR
from epstein_files.util.helpers.rich_helpers import TextCast, enclose, join_non_empty, join_texts, parenthesize

HTTPS = 'https://'
LINK_REGEX = re.compile(r"^https?://.*")
LINK_HREF_LINE_REGEX = re.compile(r"^([>• ]*)(http\S+)(.*)")
TLD_REGEX = re.compile(r"\.(com|co.uk|gov|net)$")

EXTERNAL_LINK_STYLE = 'light_slate_grey bold'
LINK_COMMENT_STYLE = 'color(195) dim italic'
OFFICIAL_LINK_STYLE = 'navajo_white3 bold'
LINK_COMMENT_PARENTHESES_STYLE = 'gray27'
SOCIAL_MEDIA_LINK_STYLE = 'pale_turquoise4'
SUBSTACK_POST_LINK_STYLE = 'bright_cyan'

SOCIAL_PLATFORMS = {
    'universeodon': 'mastodon',
    'x.com': 'twitter',
}


@dataclass
class ExternalLink(TextCast):
    """
    Container for rich `Text` links with optional parenthetical comment.
    """
    url: str
    link_text: str = ''
    comment: str = ''
    comment_url: str = ''
    comment_style: str | Style = LINK_COMMENT_STYLE
    link_style: str | Style = EXTERNAL_LINK_STYLE
    parentheses_style: str = LINK_COMMENT_PARENTHESES_STYLE

    def __post_init__(self):
        if self.comment_url and not self.comment:
            raise ValueError(f"comment_url='{self.comment_url}' but no actual comment to attach it to.")

        self.url = coerce_https(self.url)
        self.comment_url = coerce_https(self.comment_url) if self.comment_url else self.comment_url
        # TODO: shouldn't need to convert, better to work with Style objs sometimes
        self.link_style = str(self.link_style)
        self.comment_style = str(self.comment_style)

    @classmethod
    def official_link(cls, url: str, link_text: str, comment: str, comment_url: str, **kwargs) -> Self:
        """Alternate constructor for differently colored links to official sources like justice.gov."""
        def pop_with_default(arg: str, default_value: str) -> str:
            return kwargs.pop(arg) if arg in kwargs else default_value

        return cls(
            url=url,
            link_text=link_text,
            comment=comment,
            comment_style=pop_with_default('comment_style', ARCHIVE_ALT_LINK_STYLE),
            comment_url=comment_url,
            link_style=pop_with_default('link_style', OFFICIAL_LINK_STYLE),
            parentheses_style='wheat4',
            **kwargs
        )

    @classmethod
    def social_link(cls, url: str, platform_name: str = '') -> Self:
        """Alternate constructor for social media links."""
        if not platform_name:
            for domain_str, platform in SOCIAL_PLATFORMS.items():
                if domain_str in url:
                    platform_name = platform
                    break

        link = cls(url, platform_name, link_style=SOCIAL_MEDIA_LINK_STYLE)
        link.link_text = f"@{link.link_text or link.domain_stem}"
        return link

    @property
    def domain_link(self) -> Text:
        """Returns a link using the TLD free domain as the `link_text`."""
        return enclose(link_text_obj(self.url, self.domain(True), self.link_style), '[]')

    @property
    def domain_stem(self) -> str:
        """e.g. retrun 'github' for a github.com/blah URL."""
        domain_pieces = self.domain().split('.')

        if len(domain_pieces) == 2:
            return domain_pieces[0]
        elif len(domain_pieces) == 3:
            return domain_pieces[1]
        else:
            raise ValueError(f"Can't extract domain_stem for '{self.url}' (domain_pieces={domain_pieces})")

    @property
    def link(self) -> Text:
        """Link without the (comment) part."""
        return link_text_obj(self.url, self.link_text, self.link_style)

    @property
    def short_url(self) -> str:
        return self.url.removeprefix(HTTPS)

    @property
    def url_link(self) -> Text:
        """Link that uses the short_url as link_text (overriding actual link_text property)."""
        return link_text_obj(self.url, self.short_url, self.link_style)

    def domain(self, strip_tld: bool = False) -> str:
        return extract_domain(self.url, strip_tld=strip_tld)

    @classmethod
    def parenthesized_links(cls, _links: Sequence[Self | Text], base_style: str = 'white') -> Text:
        """Concatenate a collection of links and wrap in parentheses."""
        links = [link if isinstance(link, Text) else link.link for link in _links]
        links = [parenthesize(link) for link in links]
        return Text('', style=base_style).append(join_texts(links))

    def __rich__(self) -> Text:
        comment = Text('')

        if self.comment_url:
            comment = link_text_obj(self.comment_url, self.comment, self.comment_style)
            comment = enclose(comment, '()')
        elif self.comment:
            comment = enclose(Text(self.comment, self.comment_style), '()', self.parentheses_style)

        return join_non_empty(self.link, comment)


def coerce_https(url: str) -> str:
    """Prepend https:// if it's not there already."""
    return url if LINK_REGEX.match(url) else f"https://{url}"


def extract_domain(url: str, strip_tld: bool = False) -> str:
    if (domain := urlsplit(url).hostname):
        domain = domain.removeprefix('www.')
        return TLD_REGEX.sub('', domain) if strip_tld else domain
    else:
        raise ValueError(f"no hostname in URL '{url}'")


def hyperlink_line(line: str) -> Text:
    """Handles single line only. Add [link] tags if appropriate."""
    if (match := LINK_HREF_LINE_REGEX.match(line)):
        link = match.group(2)
        txt = Text(match.group(1))
        txt.append(Text.from_markup(f"[link={link}]{link}[/link]"))
        return txt.append(match.group(3))
    else:
        return Text(line)


def hyperlink_text(text: str) -> Text:
    """Add rich Text hyperlinks to a string with newlines in it."""
    return join_texts([hyperlink_line(line) for line in text.split('\n')], '\n')


def link_markup(
    url: str,
    link_text: str = '',
    style: str | None = ARCHIVE_LINK_COLOR,
    underline: bool = True
) -> str:
    """Create a rich markup string that can be turned to colored/linked `Text` with `Text.from_markup()`."""
    link_text = link_text or url.removeprefix('https://')
    style = ((style or '') + (' underline' if underline else '')).strip()
    return f"[link={url}][{style}]{link_text}[/{style}][/link]"


def link_text_obj(url: str, link_text: str = '', style: str = ARCHIVE_LINK_COLOR) -> Text:
    return Text.from_markup(link_markup(url, link_text, style))
